/* Cable and Wireless Customization of Jabber Instant Messaging Server */

/* O.D.G. Griffiths     Started: 14-Sep-00 */

/* ATTENTION: Jabber Server Version: 1.4 */

/* Source file name:    cw_ldapq.c */


/**********************/
/* LDAP Query Handler */
/**********************/
/*
 * Contains functions to acquire LDAP configuration data, to connect
 * to the LDAP database and to search for user authentication info.
 *
 * Configuration information is passed in an "mod_ldap_conf" structure,
 * defined in the "cw_sysdef.h" file.  (NOTE: This structure has been
 * cloned from the "xdb_ldap.c" source module for development purposes.)
 *
 *
 * OUTSTANDING ISSUES
 * ==================
 * 1)  ##########
 *     ##########
 */


#include "cw_sysdef.h"


/***************/
/* Global Data */
/***************/

/*** Function Templates (for forward references to local functions) ***/

int js_cw_event_mon( void* arg );
int js_cw_waitx( int rc, LDAPMessage** result, waitevt w );


/*******************/
/* Local Constants */
/*******************/

static const char* config_msg =
    "No \"%s\" attribute found in <xdb id=\"xdb_ldap\"> config section";

static const char* ldap_noconn_msg =
                        "Could not connect to LDAP Directory";

static const char* ldap_nopass_msg =
                        "Password not in LDAP Directory";

static const char* ldap_profail_msg = "LDAP processing error";


/*************************************************************************
 *  js_cw_ldap_config -- Get LDAP Connection Info from Config File.
 *
 *  Extracts the LDAP configuration parameters from the configuration
 *  file and stores them locally.
 *
 *  NOTE:  This code has been cloned from xdb_ldap() in xdb_ldap.c
 *  ====   for development purposes.
 *
 *  parameters
 *      ldap    - Pointer to the <load main="libcwldap"> node in the
 *                Config file.
 *
 *  returns
 *      If successful:  Pointer to "mod_ldap_conf" structure.  A
 *              "mod_ldap_conf" structure is constructed locally, the
 *              parameter values are loaded into it and a pointer to
 *              this structure is returned to the caller.
 *              ( Note: The parameter values are extracted direct from
 *              the Config file.)
 *      If parameter retrieval failed:  NULL.
 *
 *  NOTE:   The returned value points to a memory block that was acquired
 *  ====    via the "pmalloc" function.  The caller should not need to
 *          take any action to release it, as it should be freed
 *          automatically by the garbage collector.
 */
#define FUNCNAME    "js_cw_ldap_config"
#define VCARD       5

mod_ldap_conf js_cw_ldap_config( xmlnode ldap )
{
    mod_ldap_conf   cfg;
    char* vcard_mappings[VCARD]={"FN","GIVEN","FAMILY","NICKNAME","EMAIL"};
    int i=0;
    vcard *q;

    log_debug(FUNCNAME, "init");

    if ( j_strcmp( xmlnode_get_name( ldap ), "load" ) != 0 )
    {
        log_error( FUNCNAME, "Node passed is not <xdb>" );
        return NULL;
    }

    /* Create an "mod_ldap_conf" structure.  (Used in xdb_ldap.c to
     * hold the configuration data.) */

    /* read in vcard values - doesn't matter if the vcard values are null, 
       we just ignore it */
    
    cfg = pmalloc(xmlnode_pool(ldap),sizeof(_mod_ldap_conf));

    cfg->vcard_attrib=pmalloc(xmlnode_pool(ldap),sizeof(_vcard));
    *q=cfg->vcard_attrib;
    
    for (i=0;i<VCARD;i++)
      {
	cfg->vcard_attrib->name=vcard_mappings[i];
	cfg->vcard_attrib->mapping=xmlnode_get_attrib(ldap,vcard_mappings[i]);
	
	/* fill in config file if there is a value for mapping 
	   and create node*/
	
	if (cfg->vcard_attrib->mapping)
	  {
	    /* copy data from p to and create new node */
	    cfg->vcard_attrib->next=pmalloc(xmlnode_pool(ldap),sizeof(_vcard)); 
	    cfg->vcard_attrib=cfg->vcard_attrib->next;
	  }
      }
    /* reset what cfg->vcard_attrib is pointing to */
    cfg->vcard_attrib=*q;

    /* now extract LDAP parameters from configuration file and load
      them into the "mod_ldap_conf" structure. */

    if ( ( cfg->host = xmlnode_get_attrib(ldap, "ldap-host") )
         == NULL )
    {
        log_error(FUNCNAME, config_msg, "ldap-host");
    }
    else if ( ( cfg->port = xmlnode_get_attrib(ldap, "ldap-port") )
         == NULL )
    {
        log_error(FUNCNAME, config_msg, "ldap-port");
    }
    else if ( ( cfg->mgrdn = xmlnode_get_attrib(ldap, "ldap-mgrdn") )
         == NULL )
    {
        log_error(FUNCNAME, config_msg, "ldap-mgrdn");
    }
    else if ( ( cfg->mgrpwd = xmlnode_get_attrib(ldap, "ldap-mgrpwd") )
         == NULL )
    {
        log_error(FUNCNAME, config_msg, "ldap-mgrpwd");
    }
    else if ( ( cfg->user_attrib =
                    xmlnode_get_attrib(ldap, "ldap-user-attribute") )
         == NULL )
    {
        log_error(FUNCNAME, config_msg, "ldap-user-attribute");
    }
    else if ( ( cfg->password_attrib =
                    xmlnode_get_attrib(ldap, "ldap-password-attribute") )
         == NULL )
    {
        log_error(FUNCNAME, config_msg, "ldap-password-attribute");
    }
    else if ( ( cfg->roster_attrib =
                    xmlnode_get_attrib(ldap, "ldap-roster-attribute") )
         == NULL )
    {
        log_error(FUNCNAME, config_msg, "ldap-roster-attribute");
    }
    else if ( ( cfg->visplist =
                    xmlnode_get_attrib (ldap, "ldap-visplist") )
         == NULL )
    {
        log_error(FUNCNAME, config_msg, "ldap-visplist");
    }
    else return cfg;    /* No errors, so return the data */

    return NULL;        /* - but if any errors - return NULL */
}
#undef FUNCNAME


/*************************************************************************
 *  js_cw_ldap_init -- Establish connection to the LDAP directory and
 *                     save the connection info.
 *
 *  Opens a connection to the LDAP (X.500) directory using the access
 *  credentials defined in the configuration file and returns a pointer
 *  to the connection object for use in the subsequent authentication
 *  query.  (Note that this bind is synchronous, whereas the
 *  authentication bind in the js_cw_checkUser() function, below, is
 *  asynchronous.)
 *
 *  NOTE:  The Bind operation performed by this function uses the DN and
 *  ====   Password obtained from the <mgrdn> and <mgrpwd> elements in
 *         the configuration file, UNLIKE the Bind subsequently performed
 *         in the js_cw_checkUser() function (below), which uses a DN
 *         constructed by combining the user's name with the base string
 *         extracted from the Search Base lookup table and retrieves the
 *         password from the incoming authentication request message.
 *
 *  parameters
 *      None -- Parameters are extracted direct from the Config file.
 *
 *  returns
 *      Pointer to an LDAP if connection opened successfully.
 *      NULL if connection failed.
 */
#define FUNCNAME    "js_cw_ldap_init"

LDAP* js_cw_ldap_init( mod_ldap_conf cfg )
{
    int         rc;
    LDAP*       ld;

    /* Attempt Connection to LDAP Directory */

    log_debug( ZONE, ": " FUNCNAME "(): Getting LDAP connection");

    if ( ( ld = ldap_init(cfg->host, atoi( cfg->port ) ) ) == NULL )
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                   "Didn't get LDAP connection - "
                   "LDAP initialization failed" );
        return NULL;
    }
    else
    {
        log_debug( ZONE, ": " FUNCNAME "(): got LDAP connection: %X",
                   ld );

        /* The following line was included in Mark Wilcox's
         * original code, with the comment "prudent, but probably
         * unnecessary".  However, the "ld_deref" field appears to
         * cause a compilation error on True64, so this line is now
         * commented out here to give a clean compile.
         */
        /* ld->ld_deref = LDAP_DEREF_NEVER; */

        /* Attempt Simple (i.e. non-Kerberos) Synchronous (i.e.
         * non-blocking) bind to the LDAP connection.
         */
        log_debug( ZONE, ": " FUNCNAME "(): Attempting bind");

        if ( ( rc = ldap_simple_bind_s( ld, cfg->mgrdn, cfg->mgrpwd ) )
             == 0 )
        {
            log_debug( ZONE, ": " FUNCNAME "(): "
                       "Bind success" );  
            return ld;
        }
        else
        {
            log_error( ZONE, ": " FUNCNAME "(): "
                       "Didn't get LDAP connection - "
                       "\'simple_bind\' failed" );
            return NULL;
        }
    }
}
#undef FUNCNAME


/*************************************************************************
 *  js_cw_ldap_drop -- Release LDAP connection.
 *
 *  Releases previously opened connection to the LDAP (X.500) directory.
 *
 *  parameters
 *      cfg  -- Pointer to the "mod_ldap_conf" structure that contains
 *              the configuration data.
 *      ld   -- Pointer to the LDAP connection object.
 *
 *  returns
 *      Nothing.
 */
#define FUNCNAME    "js_cw_ldap_drop"

void js_cw_ldap_drop( mod_ldap_conf cfg, LDAP* ld )
{
    log_debug( ZONE, ": " FUNCNAME "(): Closing LDAP connection: %X",
               ld );

    if (ld != NULL)
    {
        ldap_unbind(ld);
    }

    log_debug( ZONE, ": " FUNCNAME "(): LDAP connection closed");

    return;
}
#undef FUNCNAME


/*************************************************************************
 *  js_cw_checkUser  -- Given User and Password, check against LDAP
 *                      Directory.
 *
 *  Verifies the password using the LDAP "simple bind" function.  This
 *  does not extract the password from the LDAP database, but compares
 *  the password supplied in the call against the one in the database
 *  and reports success or failure.  (The asynchronous form of the
 *  simple bind function is used, so that multiple searches may proceed
 *  in parallel.
 *
 *  NOTE:  The Bind operation performed by this function uses a DN
 *  ====   constructed by combining the user's name with the base string
 *         extracted from the Search Base lookup table and retrieves the
 *         password from the incoming authentication request message,
 *         UNLIKE the Bind performed in the js_cw_ldap_init() function,
 *         above, which uses the DN and Password obtained from the
 *         <mgrdn> and <mgrpwd> elements in the configuration file.
 *
 *  parameters
 *      dpkt     -- Pointer to the "dpacket" wrapper for the request
 *                  message.  (The "jid" in this wrapper contains both
 *                  the user-name string and the server-name string.
 *                  The wrapper itself contains a pointer to the
 *                  "xmlnode" into which the <password> tag is to be
 *                  inserted and also a pointer to its memory pool,
 *                  which should be used for this transaction.)
 *      ld       -- Pointer to LDAP connection object (returned by
 *                  "js_cw_ldap_init" when connection is established).
 *      cfg      -- Pointer to LDAP configuration data object (contains
 *                  LDAP parameters retrieved by "js_cw_ldap_config").
 *
 *  returns
 *      1 if password correct for this user.
 *      0 if password incorrect.
 */
#define FUNCNAME    "js_cw_checkUser"

result js_cw_checkUser( dpacket        dpkt,
                        LDAP*          ld,
                        mod_ldap_conf  cfg )
{
    waitevt       w;
    pth_event_t   evt;
    int           have_evt = 0;
    result        retval = r_ERR;  /* Assume failure till achieve success */
    spool         sp;
    char*         filter;
    char*         attribs[2];
    LDAPMessage*  qresult;        /* Copy of pointer to LDAP query result */
    LDAPMessage*  e = NULL;
    BerElement*   ber;
    char*         dn = NULL;
    char*         attribute;
    char**        vals   = NULL;
    xmlnode       pwtag  = NULL;
    const char*   errmsg = NULL;

    log_debug( ZONE, ": " FUNCNAME "(): Starting LDAP lookup for user" );

    /* The generic main-line for the thread handles opening of the LDAP
     * connection (via ldap_init and bind calls), but may, of course fail
     * in its attempt to do so.  We therefore detect this here by checking
     * whether the passed LDAP handle is NULL.
     */
    if ( ld == NULL)
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                   "No connection to LDAP server" );
        errmsg = ldap_noconn_msg;
        goto REPORTERR;
    }

    /* Construct the data object to be sent to the wait-event */

    w     = pmalloc( dpkt->id->p, sizeof(_waitevt) );
    w->ld = ld;

    /* construct the LDAP Bind argument by concatenating the
     * elements into a single string, then use this to initiate
     * the (non-blocking) query to the LDAP Directory.
     */

    /* Define which attribute types we want returned by the search. */

    attribs[0] = cfg->password_attrib;     /* Just the C&W password field */
    attribs[1] = NULL;            /* List terminator   */

    /* Use a string pool to construct the LDAP Bind argument by
     * concatenating the elements into a single string.
     */

    sp = spool_new( dpkt->id->p );
    spooler( sp, cfg->user_attrib, "=", dpkt->id->user, sp );
    filter = spool_print( sp );
    log_debug( ZONE, ": " FUNCNAME "(): filter: %s", filter );

    /* Now, address the query to the LDAP Directory */

    if ( ( w->msgid = ldap_search( w->ld, cfg->base,
                                   LDAP_SCOPE_SUBTREE,
                                   filter, attribs, 0 ) )
         == -1 )
    {
        /* If there was an error in the search call itself, bail out */

        log_error( ZONE, ": " FUNCNAME "(): error in LDAP search call");
        errmsg = ldap_profail_msg;
        goto REPORTERR;
    }


    /* Wait for the response to come back from the LDAP Directory... */

    evt = pth_event( PTH_EVENT_FUNC,
                     &js_cw_event_mon, (void *)w,
                     pth_time(0,500000) );
    pth_wait(evt);                          /* Block until LDAP responds */
    have_evt = 1;                           /* Note we now have an event */

    /* ...Continue here when the wait is over */

    if ( w->result == NULL )       /* An error occured during the wait */
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                   "Error occured while waiting for result" );
        pth_event_free( evt, PTH_FREE_ALL );
        errmsg = ldap_profail_msg;
        goto REPORTERR;
    }

    log_debug( ZONE, ": " FUNCNAME "(): result: 0x%08x", w->result );

    /* Check the result of the search */

    if (w->rc !=  LDAP_RES_SEARCH_ENTRY)
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                   "Error occurred during search");
        errmsg = ldap_err2string(ldap_result2error(ld,w->result,1));
        goto REPORTERR;
    }

    else
    {
        log_debug( ZONE, ": " FUNCNAME "(): "
                   "LDAP Result should be LDAP_RES_ %d", LDAP_SUCCESS );
        
        qresult = w->result;

        if ( ( e = ldap_first_entry( w->ld, qresult ) ) == NULL )
        {
            log_error( ZONE, ": " FUNCNAME "(): "
                       "No first entry from LDAP result." );
            errmsg = ldap_profail_msg;
        }
        else if ( ( dn = ldap_get_dn( w->ld, e ) ) == NULL )
        {
            log_error( ZONE, ": " FUNCNAME "(): "
                       "No DN from first entry in LDAP result." );
            errmsg = ldap_profail_msg;
        }
        else
        {
            log_debug( ZONE, ": " FUNCNAME "(): dn is %s", dn );

            /* should only be one attribute */

            if ( ( attribute = ldap_first_attribute( w->ld, e, &ber ) )
                 == NULL )
            {
                log_error( ZONE, ": " FUNCNAME "(): "
                           "No first attribute from LDAP entry." );
                errmsg = ldap_profail_msg;
            }
            else if ( ( vals = ldap_get_values( w->ld, e, attribute ) )
                      == NULL )
            {
                log_error( ZONE, ": " FUNCNAME "(): "
                           "No values from LDAP first attribute." );
                errmsg = ldap_profail_msg;
            }
            else if ( j_strlen( vals[0] ) == 0 )
            {
                log_error( ZONE, ": " FUNCNAME "(): "
                           "Password from LDAP has zero length." );
                errmsg = ldap_nopass_msg;
            }
            else if ( ( pwtag = xmlnode_insert_tag( dpkt->x, "password" ) )
                      == NULL )
            {
                log_error( ZONE, ": " FUNCNAME "(): "
                           "Could not create <password> tag xmlnode." );
                errmsg = ldap_profail_msg;
            }
            else
            {
                /* LDAP Password retrieval has succeeded, and <password>
                 * tag has been created, so insert retrieved password as
                 * data for this tag.
                 */
                xmlnode_insert_cdata( pwtag, vals[0], -1 );

                log_debug( ZONE, ": " FUNCNAME "(): "
                           "Constructed \"password\" tag:  %s",
                           xmlnode2str( pwtag ) );

                retval = r_DONE;          /* We finally made it */
            }
        }

        /* Clean up any memory usage that succeeded */

        if ( ber  != NULL )  ber_free( ber, 0 );
        if ( vals != NULL )  ldap_value_free( vals ); 
        if ( e    != NULL )  free( e );
        if ( dn   != NULL )  free( dn );
    }

REPORTERR:
    if ( retval == r_ERR )
    {
        /* If unsuccessful for any reason, attempt to insert an error
         * message in <error>...</error> tags into the message (instead
         * of the password).
         */
        if ( ( pwtag = xmlnode_insert_tag( dpkt->x, "error" ) )
                  != NULL )
        {
            xmlnode_insert_cdata( pwtag, errmsg, -1 );

            log_debug( ZONE, ": " FUNCNAME "(): "
                       "Constructed \"error\" tag:  %s",
                       xmlnode2str( pwtag ) );
        }
    }

    log_debug( ZONE, ": " FUNCNAME "(): Assessment returned %d", retval );

    /* Free up temporary resources (LDAP object is freed by caller) */

    if ( have_evt )  pth_event_free( evt, PTH_FREE_ALL );

    log_debug( ZONE, ": " FUNCNAME "(): LDAP lookup completed");

    return retval;
}
#undef FUNCNAME


/*************************************************************************
 *  js_cw_event_mon  -- Call-back function for monitoring for
 *                      completion of the LDAP Directory query.
 *  
 *  parameters
 *      arg      -- Pointer to "waitevt" structure that conveys all
 *                  dynamic information to this call-back function.
 *
 *  returns
 *      1 if we have a result (good or bad) - i.e. stop waiting.
 *      0 if still waiting.
 */
#define FUNCNAME    "js_cw_event_mon"

int js_cw_event_mon( void* arg )
{
    waitevt         w = (waitevt) arg;  /* Save data pointer locally */
    struct timeval  tv;
    LDAPMessage*    result;
    int             rc;
    int             retval;

    tv.tv_sec  = 0; /* Set time value to request non-blocking call */
    tv.tv_usec = 0;

    /* Check to see if LDAP result has been returned (but don't wait) */

    rc = ldap_result( w->ld, w->msgid, 0, &tv, &result );

    if (rc == 0)  return 0;        /* Still waiting */
    else
        retval = js_cw_waitx( rc, &result, w );

    log_debug( ZONE, ": " FUNCNAME "(): "
               "End-of-Wait function returns %d;  result ptr: 0x%08x",
               retval, result );

    /* The waiting is over!  Let the people know! */

    w->result = result; /* Store the result */
    return 1;
}
#undef FUNCNAME


/*************************************************************************
 *  js_cw_waitx  -- Callback function to check result when the wait
 *                  the wait for LDAP search completion ends.
 *  
 *  parameters
 *      rc       -- Value returned by the call to "ldap_result".
 *      result   -- Pointer to the "result" pointer argument modified
 *                  by the call to "ldap_result".
 *      w        -- Pointer to the wait event control block.
 *
 *  returns
 *      1 if successful, 0 if bind failed.
 */
#define FUNCNAME    "js_cw_waitx"

int js_cw_waitx( int rc, LDAPMessage** result, waitevt w )
{
    log_debug( ZONE, ": " FUNCNAME "(): Do end-of-wait processing");

    /* This function is only called when the LDAP Wait finishes. */

    log_debug( ZONE, ": " FUNCNAME "(): "
               "rc is %d, (remember LDAP_RES_SEARCH_ENTRY is %d)",
               rc, LDAP_RES_SEARCH_ENTRY );
 
//    log_debug( ZONE, ": " FUNCNAME "(): "
//               "After search - Result = error %d",
//               ldap_result2error( w->ld, *result, 0 ) );

    if (rc == -1) /* Error */
    {
        log_debug( ZONE, ": " FUNCNAME "(): "
                   "Result error %d, (LDAP_SUCCESS is %d)",
                   ldap_result2error( w->ld, *result, 1 ),
                   LDAP_SUCCESS );
        w->result = NULL;
        w->rc = -1;
        return 1;
    }
    else if ((rc == LDAP_RES_SEARCH_ENTRY) || (rc == LDAP_RES_SEARCH_RESULT))
    {
        log_debug( ZONE, ": " FUNCNAME "(): Returned a search entry");

        if ( ( result == NULL ) || ( *result == NULL ) )
        {
            log_debug( ZONE, ": " FUNCNAME "(): Search result is null");
        }
        else
        {
            log_debug( ZONE, ": " FUNCNAME "(): Search result is not null");
        }

        w->result = *result; /* Store the result */
        w->rc = rc;
        return 1; /* Done waiting */
    }
    else
        return 0;

    return 1;
}
#undef FUNCNAME

