/* Cable and Wireless Customization of Jabber Instant Messaging Server */

/* O.D.G. Griffiths     Started: 28-Dec-00 */

/* ATTENTION: Jabber Server Version: 1.4 */

/* Source file name:    cw_ldhnd.c */

/* modified by d hepworth - 8-Mar-01
 * adding password mod, roster set/get and vcard set/get for LDAP directory */


/***************************************************/
/* Main Module for C&W LDAP Authentication Handler */
/***************************************************/
/*
 *
 * This module is based on Jabber server versions 1.2 and 1.4, which
 * feature a radically different structure to provide better scalability.
 * With these versions, the recommended way of handling alternative
 * authentication methods, such as LDAP, is to interface them at a much
 * lower level - as an alternative 'data provider' - i.e. parallel with
 * the "xdb_file" handler that provides the 'flat file' data storage
 * (including that used for authentication in a flat-file environment).
 *
 * It is also recommended that such 'add-ons' are built as separate
 * entities that are linked in at run time (NOT integral with the main
 * Jabber server build, as was done for Phase 1).  To this end, a new
 * facility has been added to the Configuration process, so that
 * dynamically loaded add-ons can be specified through a new <load>
 * tag in the Configuration file.  The server code that handles this
 * dynamic loading uses the "dlopen()" system call to load add-on
 * modules into memory  and the and "dlsym()" system call to resolve
 * references to their initialization functions.  The initialization
 * function is responsible for registering any callback functions that
 * are to be called to carry out the handler's activities and setting
 * up any global data structure that will need to be passed to the
 * callback functions when they are invoked.
 *
 * C&W's Authentication Handler is therefore built as a separate
 * shared object library, that is only linked to the main Jabber server
 * at run time via the <load> configuration tag.
 *
 * This source file contains the cold-start initialization function and
 * the call-back functions that are registered.  (The actual LDAP access
 * is handled by lower-level functions in other modules.)
 *
 *
 * OUTSTANDING ISSUES
 * ==================
 * 1)  ##########
 *     ##########
 */


#include "cw_sysdef.h"
#include "cw_sbase.h"


/*******************/
/* Local Constants */
/*******************/

#define FILES_PRIME 613  /* international area code of Melbourne, Australia */


/***************/
/* Global Data */
/***************/

/***  Global variables ***/

/*  cw__initialized 'first-time' flag for this module */
static char cw__initialized = FALSE;

/*
 *  Global variable: js__config
 *  js__config is a pointer to the root node of configuration tree
 */
xmlnode js__config = NULL;
char *js__hostname;
xmlnode   cw_xdb_node;  /* Pointer to the configuration file's <xdb...>  */
                        /* section that loads this shared object library */

pth_mutex_t* templock;  /* Temporary mutex to ensure thread safety */

/******************************************************/
/* Local Function Templates (for forward references ) */
/******************************************************/

result cw_xdb_ldap_phandler(instance i, dpacket p, void* arg);
result cw_xdb_ldap_threadproc( void* arg, LDAP* ldx );
result cw_xdb_ldap_ticktock( void* arg );
void   cw_xdb_ldap_shutdown( void* arg );

result cw_xdb_ldap_authenticate(void* arg, LDAP* ldx);
result cw_ldap_vcard(void* arg, LDAP* ldx);
result cw_ldap_roster(void* arg, LDAP* ldx);
result cw_roster_access(dpacket p, LDAP* ldx, mod_ldap_conf config, 
			HASHTABLE cache, instance i);
result cw_vcard_access(dpacket p, LDAP* ldx, mod_ldap_conf config);

/*******************************************************************************
 * libcwldap -- 'Constructor' function for the free-standing
 *              C&W LDAP authentication module.
 *
 * This function is called when the shared object library for the C&W LDAP
 * authentication module is loaded into memory (during processing of the
 * Jabber configuration file).  It checks and processes configuration
 * information specific to this module and carries out other necessary
 * initialization, such as loading the C&W VISP Search Base Lookup Table
 * into memory, registering handler callback functions and initializing the
 * LDAP Thread Pool Manager.
 *
 * (This function is the equivalent of the "xdb_file()" function in the
 * "xdb_file.c" source code module.)
 *
 * Parameters:
 *      i   -- (instance)  Session Instance object (session-specific data).
 *      x   -- (xmlnode)   Pointer to the <load ...> node in the Configuration
 *                         file that loads this dynamic shared object library.
 */
#define FUNCNAME    "libcwldap"

void libcwldap( instance i, xmlnode x )
{
    mod_ldap_conf   cfg;
    cwldglob        globdat = NULL;

    /*    xmlnode config; */
    /*    xdbcache xc;*/
    
    if ( cw__initialized ) return;  /* Do initialization once only */

    cw__initialized = TRUE;         /* Shut the first-time-only door */

    /* The parent of the <load> node should be the <xdb id="xdb_ldap"> node
     * and the parent of that <xdb ...> node should be the <jabber> root
     * node.  So check that this is true and save the pointers to the
     * <xdb> and root nodes.  (Drop out if any failures.)
     */

    if ( ( cw_xdb_node = xmlnode_get_parent( x ) ) == NULL )
    {
        /* If we can-t find the parent node for the <load...> tag that
         * we have been passed -- so abort.
         */
        log_error( ZONE, ": " FUNCNAME "(): "
                   "In Config file, LDAP <load> node has no parent" );
    }
    else if ( j_strcmp( xmlnode_get_name( cw_xdb_node ), "xdb" ) != 0 )
    {
        /* If the parent of the <load...> tag is NOT the <xdb...> tag
         * that has "xdb_ldap" as its ID -- so abort.
         */
        log_error( ZONE, ": " FUNCNAME "(): "
                   "In Config file, parent of LDAP <load> node is "
                   "not <xdb>" );
    }
    else if ( ( js__config = xmlnode_get_parent( cw_xdb_node ) ) == NULL )
    {
        /* We've found the <xdb...> node and it should have the <jabber>
         * top-level tag as its parent -- abort, if we can't find parent.
         */
        log_error( ZONE, ": " FUNCNAME "(): "
                   "In Config file, LDAP <xdb> node has no parent" );
    }
    else if ( j_strcmp( xmlnode_get_name( js__config ), "jabber" ) != 0 )
    {
        /* We found the parent, but it isn't the <jabber> tag -- so abort.
         */
        log_error( ZONE, ": " FUNCNAME "(): "
                   "In Config file, parent of LDAP <xdb> node is "
                   "not <jabber>" );
    }
    else
    do /* ... while (0); - Once-through block to permit use of break */
    {
        /*** <xdb> and Root Nodes Retrieved Successfully ***/

        log_debug( ZONE, ": " FUNCNAME "(): "
                   "C&W LDAP component initialization");

        /*** Retrieve LDAP Parameters from Configuration File ***/
        /*
         * Quit if failed to get any parameters.
         * ("js_cw_ldap_config()" will already have logged the error.)
         * Note:  "js_cw_ldap_config()" uses the memory pool of the
         * "x" parameter to construct the returned "cfg" object and
         * the data strings associated with it.  (I.e. it uses the
         * pool of the Configuration file for this.)
         */
        if ( ( cfg = js_cw_ldap_config( x ) ) == NULL )
            break;                     /* Failed */


        /*** Read in VISP Search Base Lookup Table ***/

        if ( ! js_visplist_exist() )
            if ( ! js_visplist_load( cw_xdb_node, cfg->visplist ) )
                break;                 /* Failed */


        /*** Construct the Global Data Structure for this Handler ***/

        /* When our  "cw_xdb_ldap_phandler" callback function is called,
         * it will be passed a pointer to this structure (which is
         * therefore created as persistent storage, in the pool of the
         * current instance).
         */

        globdat = pmalloco( i->p, sizeof( _cwldglob ) );
        globdat->i = i;
        globdat->cfg = cfg;
	globdat->timeout=-1; /* defaults to timeout forever */
	globdat->cache=ghash_create(FILES_PRIME,(KEYHASHFUNC)str_hash_code,
				    (KEYCOMPAREFUNC)j_strcmp);


        /*** Initialize the Thread Pool Manager ***/

        ldap_threadman_init( cfg );

        /* Also initialize the mutex used to make "cw_xdb_ldap_phandler"
         * (which accesses the thread pool) thread-safe.
         */
        pth_mutex_init( templock );


        /*** All preliminaries OK - So Register Handlers ***/

        /* Register the callback function to handle LDAP authentication
         * requests.
         * Parameters for "register_phandler()" are:
         *   1) ("instance")  Instance ident, as passed to THIS function.
         *   2) ("order")     Handler priority - "o_DELIVER" for "xdb".
         *   3) ("phandler")  Pointer to handler's callback function.
         *   4) ("void *")    Pointer to this handler's global data.
         */
        register_phandler( i, o_DELIVER, cw_xdb_ldap_phandler,
                           (void *)globdat);

        // Register the callback function to handle regular refresh and
        // tidy-up activities.
        // Parameters for "register_beat()" are:
        //   1) ("freq")         Calling frequency (in seconds?)
        //   2) ("beathandler")  Pointer to handler's callback function.
        //   3) ("void *")       Pointer to this handler's global data.

        register_beat( 30, cw_xdb_ldap_ticktock, (void *)globdat );

        /* Register the callback function to close LDAP connections,
         * threads, etc. at program shutdown.
         */
        register_shutdown( cw_xdb_ldap_shutdown, (void *)globdat );

        return;         /* Return from successful setup */
    }
    while (0);

    /* If we reach here, we have failed, so register 'killer' handlers. */


        // #################### TO BE COMPLETED ####################



        // ########## REGISTER KILLER HANDLER FOR ZERO-K ###########

}
#undef FUNCNAME


/*******************************************************************************
 * cw_xdb_ldap_phandler -- Callback function for handling Authentication
 *                         reqests.
 *
 * This function is called each time a request is made for data that is
 * within the group of namespaces assigned to the CW LDAP data handler.
 * (The namespaces are defined via <ns> tags in the Configuration file.)
 *
 * Parameters:
 *      i    -- Pointer to the "instance" control block for the current
 *              instance.
 *      p    -- Pointer to the "dpacket" containing the request message.
 *      arg  -- Pointer to the parameter block for this callback function
 *              (constructed by "libcwldap" 'constructor' function, above).
 */
#define FUNCNAME    "cw_xdb_ldap_phandler"

result cw_xdb_ldap_phandler(instance i, dpacket p, void* arg)
{
    int       outcome = r_DONE;        /* Assume OK until proved otherwise */
    cwldglob  globdat = (cwldglob)arg;   /* CW LDAP static global data */
    ldapdata thrdat  = NULL;

    /* Make this function thread-safe (and any functions that it calls)
     * by wrapping a mutex around it so that calls can only invoke it
     * one at a time.  This should not affect performance significantly
     * because nothing in this function (or the ones that it calls) are
     * time-consuming.  The delays come later, when a separate thread
     * has been allocated.
     */
    pth_mutex_acquire( templock, FALSE, NULL );

    /*#ifdef NEVER*/
    
    /* now checking for certain types of requests, namely
     * authentication, vcard and roster requests. these requests can
     * be of type set or get.
     * 
     * all of these checks are 'instantaneous' - i.e. they
     * are all in-memory on the local machine.  
     *
     * if these checks are successful, then we now have to 
     * initiate LDAP transactions which involve the remote X.500 
     * server, so these need to be run on a thread of their own 
     * to allow other LDAP queries to proceed in parallel. 
     * so first construct the data block that is to be passed to the 
     * callback function, then pass this to the function that allocates 
     * threads, along with the name of the callback function.
     */

    log_debug( ZONE, ": " FUNCNAME "(): "
               "C&W LDAP handling request:  %s",
               xmlnode2str( p->x ) );
    
    
    /* checking for valid VISP name */

    if ((globdat->cfg->base = js_vispbase(p->id->server))==NULL)
    {
        /* the server name extracted from the packet's "jid" field must be
         * that of a recognized VISP -- abort if not.
         */
        log_error( ZONE, ": " FUNCNAME "(): "
                   "VISP name not defined in look-up table." );
        outcome = r_ERR;
    }


    /* check for required user name */
    
    else if ( p->id->user == NULL )
      {
        /* The required User must be specified via the packet's "jid" field
         * -- abort if not.  (No User implies a request for Global data.)
         */
        log_error( ZONE, ": " FUNCNAME "(): "
                   "Attempt to reference global data via LDAP "
                   "authentication.  LDAP only valid for specific users." );
        outcome = r_ERR;
    }


    /* check for authentication request */

    else if (j_strcmp(xmlnode_get_attrib(p->x,"ns"),"jabber:iq:auth")==0)
      {
	/* check to see if authentication request */

	if (j_strcmp(xmlnode_get_attrib(p->x,"type"),"get")==0)
	  {
	    /*	    log_debug( ZONE, ": " FUNCNAME "(): "
		       "C&W LDAP authentication request:  %s",
		       xmlnode2str( p->x ) );*/
	    
	    thrdat          = pmalloco( p->p, sizeof( _ldapdata ) );
	    thrdat->globdat = globdat;
	    thrdat->dpkt    = p;
	    thrdat->inst    = i;
	    
	    ldap_threadman_send( p->p, cw_xdb_ldap_authenticate, (void*)thrdat );
	  }

	/* check to see if password change request */

	else if (j_strcmp(xmlnode_get_attrib(p->x,"type"),"set")==0)
	  {
	    log_error( ZONE, ": " FUNCNAME "(): "
		       "Password change not allowed." );
	    outcome = r_ERR;
	  }
      }
    

    /* check for roster access */

    else if (j_strcmp(xmlnode_get_attrib(p->x,"ns"),"jabber:iq:roster")==0)
      {
	thrdat = pmalloco( p->p, sizeof( _ldapdata ) );
	thrdat->globdat = globdat;
	thrdat->dpkt = p;
	thrdat->inst = i;
	
	ldap_threadman_send( p->p, cw_ldap_roster, (void*)thrdat );
      }

    /* check for vcard access */
    
    else if (j_strcmp(xmlnode_get_attrib(p->x,"ns"),"vcard-temp")==0)
      {
	thrdat = pmalloco( p->p, sizeof( _ldapdata ) );
	thrdat->globdat = globdat;
	thrdat->dpkt = p;
	thrdat->inst = i;
	
	ldap_threadman_send( p->p, cw_ldap_vcard, (void*)thrdat );
      }

    /* From here on, events are beyond this function's control.  If an
     * error occurs in the callback function, it can't reported back here
     * as this function will be long gone.  So the callback function will
     * have to report errors by constructing an error message as its
     * response if this is appropriate.
     */
    
    pth_mutex_release( templock );     /* Release the thread-safe lock. */

    return outcome;
}
#undef FUNCNAME




/*******************************************************************************
 * cw_xdb_ldap_ticktock -- Callback function for tidy-up actions at regular
 *                         time intervals.
 *
 * Checks for and deals with requests that have timed out, etc.
 *
 * Parameters:
 *      arg  -- Pointer to the parameter block for this callback function
 *              (constructed by "libcwldap" 'constructor' function, above).
 */
#define FUNCNAME    "cw_xdb_ldap_ticktock"

result cw_xdb_ldap_ticktock( void* arg )
{
    int   outcome = r_DONE;        /* Assume OK until proved otherwise */


    // ########### NO REGULAR ACTIONS AT THIS TIME ############

    // ############### ANY UNFINISHED BUSINESS? ###############


    return outcome;
}
#undef FUNCNAME


/*******************************************************************************
 * cw_xdb_ldap_shutdown -- Callback function for clean-up at program
 *                         termination.
 *
 * Closes down threads, LDAP connections, etc., at instance shutdown.
 *
 * Parameters:
 *      arg  -- Pointer to the parameter block for this callback function
 *              (constructed by "libcwldap" 'constructor' function, above).
 */
#define FUNCNAME    "cw_xdb_ldap_shutdown"

void cw_xdb_ldap_shutdown( void* arg )
{
    cwldglob  globdat = (cwldglob)arg;

    /*** Shut Down the Thread Pool Manager ***/

    ldap_threadman_shutdown( globdat->cfg );
}
#undef FUNCNAME


/*******************************************************************************
 * cw_xdb_ldap_authenticate -- Callback function for handling multi-threaded
 *                             LDAP authentication actions.
 *
 * This function is passed to the LDAP thread pool manager by
 * "cw_xdb_ldap_phandler", above.  It retrieves a user's password from the
 * X.500 directory.
 *
 * Parameters:
 *      arg  -- Pointer to the (ldapdata) parameter block for this
 *              function.  (Since the pointer has to be passed via the
 *              thread pool manager which will ultimately handle more
 *              than one kind of callback function, the pointer has to
 *              be passed as a pointer-to-void.)
 *      ld   -- Pointer to the LDAP connection for the assigned thread.
 *              (For thread-safety, each thread needs its own LDAP
 *              connection.  In the interests of efficiency, each thread
 *              keeps its LDAP connection open between one LDAP query
 *              and the next, avoiding the need for an initial "bind"
 *              operation every time.  The LDAP connection is therefore
 *              opened in the generic main-line for CW LDAP threads and
 *              passed through to the query-specific callback function
 *              here.)
 */
#define FUNCNAME    "cw_xdb_ldap_authenticate"

result cw_xdb_ldap_authenticate( void* arg, LDAP* ld )
{
    ldapdata   thrdat = (ldapdata)arg;  /* Thread-process-specific data */
    result      retval;
    const char* typetext;

    log_debug( ZONE, ": " FUNCNAME "(): Using LDAP connection: %X",
               ld );

    log_debug( ZONE, ": " FUNCNAME "(): User ident: %s@%s/%s",
               thrdat->dpkt->id->user,
               thrdat->dpkt->id->server,
               thrdat->dpkt->id->resource );

    /* Obtain the password for this user from the LDAP directory. If
     * the retrieval is successful, "js_cw_checkUser()" will insert a
     * <password> tag (with the extracted password as its data) into
     * the xmlnode that is wrapped by the "dpacket".
     * If unsuccessful, "js_cw_checkUser()" will insert an <error> tag
     * with a message indicating the type of error.
     */
    retval = js_cw_checkUser( thrdat->dpkt, ld, thrdat->globdat->cfg );

    typetext = ( retval == r_DONE ) ? "result" : "error";

    /* We have to produce the same response as would be produced by the
     * flat-file function "xdb_file_phandler()", which adapts the xmlnode
     * in the incoming packet to turn it into the required response and
     * then constructs a new "dpacket" wrapper to deliver it back to
     * the original requester.
     * "js_cw_checkUser" constructs a new node as a child of the incoming
     * xdb message.  If successful, it makes the name of this node
     * <password> and fills it in with the retrieved value.  If
     * unsuccessful, it makes the name of this node <error> and fills it
     * in with an appropriate error text.
     * So a successful response appears as:
     * 
     * <xdb type="result" to="sessions"
     *      from="username@servername" ns="jabber:iq:auth" id="1">
     *   <password>plainpassword</password>
     * </xdb>
     * 
     * If "js_cw_checkUser()" fails, the returned result will be:
     * 
     * <xdb type="error" to="sessions"
     *      from="username@servername" ns="jabber:iq:auth" id="1">
     *   <error>Text of error description here</error>
     * </xdb>
     */

    xmlnode_put_attrib( thrdat->dpkt->x, "type", typetext );
    xmlnode_put_attrib( thrdat->dpkt->x, "to",
                        xmlnode_get_attrib( thrdat->dpkt->x, "from" ) );
    xmlnode_put_attrib( thrdat->dpkt->x, "from",
                        jid_full( thrdat->dpkt->id ) );
    deliver( dpacket_new( thrdat->dpkt->x ), NULL );

    return retval;
}
#undef FUNCNAME



/*******************************************************************************
 * cw_ldap_roster -- Callback function for handling multi-threaded
 *                   LDAP roster access actions.
 *
 * This function is passed to the LDAP thread pool manager by
 * "cw_xdb_ldap_phandler", above.  It retrieves a user's roster from the
 * X.500 directory.
 *
 * Parameters:
 *      arg  -- Pointer to the (ldapdata) parameter block for this
 *              function.  (Since the pointer has to be passed via the
 *              thread pool manager which will ultimately handle more
 *              than one kind of callback function, the pointer has to
 *              be passed as a pointer-to-void.)
 *      ld   -- Pointer to the LDAP connection for the assigned thread.
 *              (For thread-safety, each thread needs its own LDAP
 *              connection.  In the interests of efficiency, each thread
 *              keeps its LDAP connection open between one LDAP query
 *              and the next, avoiding the need for an initial "bind"
 *              operation every time.  The LDAP connection is therefore
 *              opened in the generic main-line for CW LDAP threads and
 *              passed through to the query-specific callback function
 *              here.)
 */
#define FUNCNAME    "cw_ldap_roster"

result cw_ldap_roster(void* arg, LDAP* ld)
{
  ldapdata thrdat = (ldapdata)arg;  /* Thread-process-specific data */
  result retval;
  
  log_debug( ZONE, ": " FUNCNAME "(): Using LDAP connection: %X",
	     ld );
  
  log_debug( ZONE, ": " FUNCNAME "(): User ident: %s@%s/%s",
	     thrdat->dpkt->id->user,
	     thrdat->dpkt->id->server,
	     thrdat->dpkt->id->resource );
  
  retval=cw_roster_access(thrdat->dpkt,ld,
			  thrdat->globdat->cfg,
			  thrdat->globdat->cache,
			  thrdat->globdat->i);
  
  
  return retval;
}
#undef FUNCNAME



/*******************************************************************************
 * cw_ldap_vcard -- Callback function for handling multi-threaded
 *                   LDAP vcard access actions.
 *
 * This function is passed to the LDAP thread pool manager by
 * "cw_xdb_ldap_phandler", above.  It retrieves a user's vcard from the
 * X.500 directory.
 *
 * Parameters:
 *      arg  -- Pointer to the (ldapdata) parameter block for this
 *              function.  (Since the pointer has to be passed via the
 *              thread pool manager which will ultimately handle more
 *              than one kind of callback function, the pointer has to
 *              be passed as a pointer-to-void.)
 *      ld   -- Pointer to the LDAP connection for the assigned thread.
 *              (For thread-safety, each thread needs its own LDAP
 *              connection.  In the interests of efficiency, each thread
 *              keeps its LDAP connection open between one LDAP query
 *              and the next, avoiding the need for an initial "bind"
 *              operation every time.  The LDAP connection is therefore
 *              opened in the generic main-line for CW LDAP threads and
 *              passed through to the query-specific callback function
 *              here.)
 */
#define FUNCNAME    "cw_ldap_vcard"

result cw_ldap_vcard(void* arg, LDAP* ld)
{
  ldapdata thrdat = (ldapdata)arg;  /* Thread-process-specific data */
  result retval;
  
  log_debug( ZONE, ": " FUNCNAME "(): Using LDAP connection: %X",
	     ld );
  
  log_debug( ZONE, ": " FUNCNAME "(): User ident: %s@%s/%s",
	     thrdat->dpkt->id->user,
	     thrdat->dpkt->id->server,
	     thrdat->dpkt->id->resource );
  
  retval=cw_vcard_access(thrdat->dpkt,ld,thrdat->globdat->cfg);
  return retval;
}
#undef FUNCNAME





