/* Cable and Wireless Jabber modification code - for roster access */


#include "cw_sysdef.h"


/***************/
/* Global Data */
/***************/

/*** Function Templates (for forward references to local functions) ***/

int js_cw_event_mon( void* arg );
int js_cw_waitx( int rc, LDAPMessage** result, waitevt w );
xmlnode cw_ldap_get_cache(const char* username, HASHTABLE cache);
char *cw_ldap_full(pool p, char *host, char *user, char *ext);
xmlnode cw_ldap_get_directory_data(dpacket p, LDAP* ld, mod_ldap_conf config,
				   result* ldap_error);
xmlnode cw_ldap_get_data (dpacket p, char *username, LDAP *ld, mod_ldap_conf config, HASHTABLE cache);
xmlnode cw_ldap_roster_get(dpacket p, LDAP* ld, mod_ldap_conf cfg);
result cw_ldap_roster_set(dpacket p, LDAP* ld, mod_ldap_conf cfg, xmlnode data);

/*******************/
/* Local Constants */
/*******************/



/* this function is used to access the LDAP directory and set or get roster information */

/* parameters - dpacket p - dpacket to the session
 *              LDAP* ld - ldap connection configuration info
 *              mod_ldap_conf config - configuration info for the directory/schema
 *         	HASHTABLE cache - cache for LDAP roster
 *
 * result - returns success or not
 */

result cw_roster_access (dpacket p, LDAP* ld, mod_ldap_conf config,
			 HASHTABLE cache, instance i)
{
  char *full, *ns, *act, *match;
  xmlnode file=NULL, top=NULL, data=NULL;
  int ret=0, flag_set=0;

  log_debug(ZONE,"handling LDAP request %s",xmlnode2str(p->x));
  
  if((ns = xmlnode_get_attrib(p->x,"ns")) == NULL)
    return r_ERR;
  
  if(j_strcmp(xmlnode_get_attrib(p->x,"type"), "set") == 0)
    flag_set = 1;
  
  /* is this request specific to a user or global data? 
     use that for the file name */

  if (p->id->user != NULL)
    full = cw_ldap_full(p->p,p->id->server,p->id->user,"xml");
  else
    full = cw_ldap_full(p->p,p->id->server,"global","xdb");
  
  if(full == NULL)
    return r_ERR;
  
  /* load the data from directory/cache */
  top = file = cw_ldap_get_data(p,full,ld,config,cache);
  
  /* if we're dealing w/ a resource, just get that element */
  if(p->id->resource != NULL)
    {
      if((top = xmlnode_get_tag(top,spools(p->p,"res?id=",p->id->resource,p->p))) == NULL)
        {
	  top = xmlnode_insert_tag(file,"res");
	  xmlnode_put_attrib(top,"id",p->id->resource);
        }
    }
  
  /* just query the relevant namespace */
  data = xmlnode_get_tag(top,spools(p->p,"?xdbns=",ns,p->p));
  
  if(flag_set)
    {
      act = xmlnode_get_attrib(p->x,"action");
      match = xmlnode_get_attrib(p->x,"match");
      if(act != NULL)
        {
	  switch(*act)
            {
            case 'i':
	      if(data == NULL)
                { /* we're inserting into something that doesn't exist?!?!? */
		  data = xmlnode_insert_tag(top,"foo");
		  xmlnode_put_attrib(data,"xdbns",ns);
		  xmlnode_put_attrib(data,"xmlns",ns); /* should have a top-level xmlns attrib */
                }
	      xmlnode_hide(xmlnode_get_tag(data,match)); /* any match is a goner */
	      /* insert the new chunk into the existing data */
	      xmlnode_insert_tag_node(data, xmlnode_get_firstchild(p->x));
	      break;
            default:
	      log_warn("xdb_file","unable to handle unknown LDAP action '%s'",act);
	      return r_ERR;
            }
        }
      else
	{
	  if(data != NULL)
	    xmlnode_hide(data);
	  
	  /* copy the new data into file */
	  data = xmlnode_insert_tag_node(top, xmlnode_get_firstchild(p->x));
	  xmlnode_put_attrib(data,"xdbns",ns);
        }
      
      /* write to directory */
      if(cw_ldap_roster_set(p,ld,config,file)!=r_ERR)
	ret=1;
      else
	log_error(p->id->server,"LDAP request failed, unable to write roster");
    }
  else
    {
      /* a get always returns, data or not */
      ret = 1;
      
      if(data != NULL)
	{ /* cool, send them back a copy of the data */
	  xmlnode_hide_attrib(xmlnode_insert_tag_node(p->x, data),"xdbns");
	}
    }
  
  if(ret)
    {
      xmlnode_put_attrib(p->x,"type","result");
      xmlnode_put_attrib(p->x,"to",xmlnode_get_attrib(p->x,"from"));
      xmlnode_put_attrib(p->x,"from",jid_full(p->id));
      deliver(dpacket_new(p->x), NULL); /* dpacket_new() shouldn't ever return NULL */

      /* remove the cached item if it was a set */
      if(flag_set)
	{
	  /*#ifdef NEVER*/
	  log_debug(ZONE,"decaching %s",full);
	  ghash_remove(cache,full);
	  /*#endif*/
	  xmlnode_free(file);
	}
      return r_DONE;
    }
  else
    {
      return r_ERR;
    }
}



/* this function is similar to jidfull, and returns user@host for caching purposes */

char *cw_ldap_full(pool p, char *host, char *user, char *ext)
{
  spool sp = spool_new(p);
  char *ret;
  
  /* full user name */
  spooler(sp,user,"@",host,"/",ext,sp);
  ret = spool_print(sp);
  
  return ret;
}



/* this function is used to get data from the cache or ldap directory */

xmlnode cw_ldap_get_data (dpacket p,char *username, LDAP *ld, mod_ldap_conf config, HASHTABLE cache)
{
  xmlnode data=NULL;
  cacher c=NULL;
  
  log_debug(ZONE,"loading %s",username);
  
  /*#ifdef NEVER  */
  /* check the cache */
  if ((c=ghash_get(cache,username))!=NULL)
    return xmlnode_str(c->file,strlen(c->file));
  /*#endif  */
  /* no data in cache, go to LDAP directory */
  data=cw_ldap_roster_get(p,ld,config);
  
  /* if LDAP directory empty, create new cache entry */
  if (data==NULL)
    {
      data=xmlnode_new_tag("xdb");
    }
  
  /*#ifdef NEVER  */
  log_debug(ZONE,"caching %s",username);
  c=pmalloco(xmlnode_pool(data),sizeof(_cacher));
  c->name=username;
  c->lastset=time(NULL);
  c->file=xmlnode2str(data);
  ghash_put(cache,c->name,c);
  /*#endif  */
  return data;
}




/* this function is to get roster information from the directory - 
 * it is modeled on the authentication function
 */

#define FUNCNAME "cw_ldap_roster_get"

xmlnode cw_ldap_roster_get(dpacket p, LDAP* ld, mod_ldap_conf cfg)
{
  xmlnode data=NULL; /* data from LDAP directory to be returned */

  waitevt w;
  pth_event_t evt;
  int have_evt=FALSE;
  spool sp;
  char *filter,*attribs[2], *dn=NULL, *attribute, *error;
  char **vals=NULL;
  LDAPMessage* qresult;
  LDAPMessage* e=NULL;
  BerElement *ber;

  log_debug(ZONE,": " FUNCNAME " (): Started LDAP roster retrieval for user");
  
  /* checking for connection to LDAP server */
  
  if (ld==NULL)
    {
      /* no connection, return NULL data */
      log_error(ZONE,": " FUNCNAME "(): No connection to LDAP server");
      return data;
    }

  /* create data object to be sent to wait-event */
  
  w=pmalloc(p->id->p,sizeof(_waitevt));
  w->ld=ld;

  /* construct LDAP binding argument and define what attributes are to 
     be returned from the search */

  attribs[0]=cfg->roster_attrib;
  attribs[1]=NULL;
  
  sp=spool_new(p->id->p);
  spooler(sp,cfg->user_attrib,"=",p->id->user,sp);
  filter=spool_print(sp);
  
  log_debug(ZONE,": " FUNCNAME "(): filter: %s",filter);
  
  /* address query to the LDAP directory */
  
  if ((w->msgid=ldap_search(w->ld,cfg->base,LDAP_SCOPE_SUBTREE,
			    filter,attribs,0))
      ==-1)
    {
      /* error in the search call */
      error=ldap_err2string(ldap_result2error(ld,w->result,1));
      log_error(ZONE,": " FUNCNAME "(): LDAP result \"%s\" occured during roster retrieve operation",error);
      /*      log_error(ZONE,": " FUNCNAME "(): error in LDAP search call");*/
      return data;
    }
  
  /* wait for response to come back from LDAP directory */
  evt = pth_event(PTH_EVENT_FUNC,&js_cw_event_mon,(void *)w,
		  pth_time(0,500000));
  pth_wait(evt); /* block until LDAP responds */
  have_evt=TRUE;

  if (w->result==NULL)
    {
      /* an error occured during wait */
      log_error(ZONE,": " FUNCNAME "(): Error occured while waiting for result");
      pth_event_free(evt,PTH_FREE_ALL);
      return data;
    }
  
  log_debug(ZONE,": " FUNCNAME "(): result: 0x%08x", w->result);

  /* check the result of the search */
  if (w->rc!= LDAP_RES_SEARCH_ENTRY)
    {
      /* add extra checking here to distinguish the LDAP error */
      error=ldap_err2string(ldap_result2error(ld,w->result,1));
      log_error(ZONE,": " FUNCNAME "(): LDAP result \"%s\" occured during roster retrieve operation",error);
    }
  
  else
    {
      log_debug(ZONE,": " FUNCNAME "(): LDAP result should be LDAP_RES_ %d",LDAP_SUCCESS);
      qresult=w->result;

      if ((e=ldap_first_entry(w->ld,qresult))==NULL)
	{
	  log_error(ZONE,": " FUNCNAME "(): No first entry from LDAP result");
	}
      else if ((dn=ldap_get_dn(w->ld,e))==NULL)
	{
	  log_error(ZONE,": " FUNCNAME "(): No dn from first entry in LDAP result");
	}
      else
	{
	  log_debug(ZONE,": " FUNCNAME "(): dn is %s",dn);
	  
	  /* should only be one attribute */
	  
	  if ((attribute=ldap_first_attribute(w->ld,e,&ber))==NULL)
	    {
	      log_error(ZONE,": " FUNCNAME "(): No first attribute from LDAP result");
	    }
	  else if ((vals=ldap_get_values(w->ld,e,attribute))==NULL)
	    {
	      log_error(ZONE,": " FUNCNAME "(): No value from LDAP first attribute");
	    }
	  else
	    {
	      /* data has been retrieved from LDAP server */
	      data=xmlnode_str(vals[0],strlen(vals[0]));
	    }
	}
      if (ber!=NULL)
	ber_free(ber,0);
      if (vals!=NULL)
	ldap_value_free(vals);
      if (e!=NULL)
	free(e);
      if (dn!=NULL)
	free(dn);
    }
  
  if (have_evt)
    pth_event_free(evt,PTH_FREE_ALL);
  
  log_debug(ZONE,": " FUNCNAME "():LDAP roster access completed");
  
  return data;
}
#undef FUNCNAME




/* this function is used to write data to the ldap directory */

#define FUNCNAME "cw_ldap_roster_set"

result cw_ldap_roster_set (dpacket p, LDAP* ld, mod_ldap_conf cfg, xmlnode data)
{
  waitevt w;
  pth_event_t evt;
  int have_evt=FALSE;
  spool sp;
  char *filter,*attribs[2], *error;
  LDAPMod modified_data,*modified[2];
  result retval=r_ERR;

  log_debug(ZONE,": " FUNCNAME " (): Started LDAP roster modify for user");
  
  /* checking for connection to LDAP server */
  
  if (ld==NULL)
    {
      /* no connection, return NULL data */
      log_error(ZONE,": " FUNCNAME "(): No connection to LDAP server");
      return retval;
    }
  
  /* create data object to be sent to wait-event */
  
  w=pmalloc(p->id->p,sizeof(_waitevt));
  w->ld=ld;
  
  /* construct LDAP binding argument and define what attributes are to 
     be sent to the modify operation */
  
  attribs[0]=xmlnode2str(data);
  attribs[1]=NULL;
 
  modified_data.mod_op=LDAP_MOD_REPLACE;
  modified_data.mod_type=cfg->roster_attrib;
  modified_data.mod_values=attribs;
  
  modified[0]=&modified_data;
  modified[1]=NULL;

  sp=spool_new(p->id->p);
  spooler(sp,cfg->user_attrib,"=",p->id->user,",",cfg->base,sp);
  filter=spool_print(sp);
  
  log_debug(ZONE,": " FUNCNAME "(): filter: %s",filter);
  
  /* address query to the LDAP directory */
  
  if ((w->msgid=ldap_modify(w->ld,filter,modified))
      ==-1)
    {
      /* error in the modify call */
      log_error(ZONE,": " FUNCNAME "(): error in LDAP modify call");
      return retval;
    }
  
  /* wait for response to come back from LDAP directory */
  evt = pth_event(PTH_EVENT_FUNC,&js_cw_event_mon,(void *)w,
		  pth_time(0,500000));
  pth_wait(evt); /* block until LDAP responds */
  have_evt=TRUE;
  
  if (w->result==NULL)
    {
      /* an error occured during wait */
      log_error(ZONE,": " FUNCNAME "(): Error occured while waiting for result");
      pth_event_free(evt,PTH_FREE_ALL);
      return retval;
    }
  
  log_debug(ZONE,": " FUNCNAME "(): result: 0x%08x", w->result);
  
  /* check the result of the modify */
  if (w->rc!= LDAP_RES_SEARCH_ENTRY)
    {
      /* add extra checking here to distinguish the LDAP result */
      error=ldap_err2string(ldap_result2error(ld,w->result,1));
      if (!j_strcmp(error,"Success"))
	{
	  /* success!!! */
	  retval=r_DONE;
	}
      else
	{
	  /* failure */
	  log_error(ZONE,": " FUNCNAME "(): LDAP result \"%s\" occured during roster modification operation",error);
	}
    }
  
  if (have_evt)
    pth_event_free(evt,PTH_FREE_ALL);
  
  log_debug(ZONE,": " FUNCNAME "():LDAP roster access completed");

  return retval;
}
#undef FUNCNAME








