/* Cable and Wireless Customization of Jabber Instant Messaging Server */

/* O.D.G. Griffiths     Started: 16-Aug-00 */

/* ATTENTION: Jabber Server Version: 1.4 */

/* Source file name:    cw_sbase.c */


/************************************/
/* Search Base Table Lookup Handler */
/************************************/
/*
 * The original Jabber Server only handles a single domain per server.
 * CW's use of Jabber must permit communication from the clients of one
 * VISP to the clients of another VISP.  Each VISP has its own domain,
 * but for Jabber's use these are all tied together in a single tree.
 * Searching this tree right from the top to locate a user would entail
 * a multi-level search across all VISP which would be unacceptably
 * slow.
 *
 * The CW implementation therefore inserts a preliminary step in which
 * the domain specified in any request from a client is first used to
 * connect to the right lower-level LDAP directory for the specified
 * VISP, so that the search can start from a base-point in the LDAP tree
 * that embraces only the users of the target VISP.  This is done via a
 * small lookup table that contains one entry per VISP [domain]
 * and returns the LDAP elements needed to constrain the search to the
 * required sub-level.
 *
 * The look-up table is structured as an XML file and use is made of
 * Jabber's existing utility functions for processing XML structures.
 * The format of the look-up table is as follows:
 *
 *     <cwimvisplist>
 *         <vbases>
 *             <visp>
 *                 <name>FirstVispName</name>
 *                 <full>firstvispname.domain1.net</full>
 *                 <base>ou=users,...ou=domains,o=cwc</base>
 *             </visp>
 *             <visp>
 *                 <name>SecondVispName</name>
 *                 <full>secondvispname.domain2.net</full>
 *                 <base>ou=users,...ou=domains,o=cwc</base>
 *             </visp>
 *             ...
 *         </vbases>
 *     </cwimvisplist>
 *
 * The contents of the lookup table only needs to change relatively
 * infrequently, when a VISP is added or deleted, or when a VISP's
 * domain identification changes.
 *
 * This module contains the functions that set up, maintain and access
 * the lookup table.
 *
 * OUTSTANDING ISSUES
 * ==================
 * 1)  Have not considered any actions needed when table is updated,
 *     rather than simply loaded at cold start.
 */


#include "cw_sysdef.h"
#include "cw_sbase.h"


/***************/
/* Global Data */
/***************/

#define CW_VBASES "vbases"  /* Tag of XML section containing lookup */
#define CW_VISP   "visp"    /* Tag of XML node for individual VISP  */
#define CW_VNAME  "name"    /* Tag of <name> element of <visp> node */
#define CW_VFULL  "full"    /* Tag of <full> element of <visp> node */
#define CW_VBASE  "base"    /* Tag of <base> element of <visp> node */

/***  Global variables ***/

/*  js__visplist points to the root node of the Lookup Table tree */
static xmlnode js__visplist = NULL;

/*  js__vispbases points to the <bases> node in the Lookup Table tree */
static xmlnode js__vispbases = NULL;

/*  js__vispfirst points to first <visp> node within the <bases> node */
static xmlnode js__vispfirst = NULL;


/*************************************************************************
 *  js_visp_norm -- Normalize a string by removing leading and trailing
 *                  whitespace and converting to all lowercase.
 *
 *  Does not alter the original string, but uses pool storage to return
 *  the converted version.
 *
 *  parameters
 *      oldtext -- Pointer to string to be normalized.
 *      pp      -- Pointer to memory pool.
 *
 *  returns
 *      Pointer to the new string (stored in memory acquired from the
 *      pool).
 */
#define FUNCNAME    "js_visp_norm"

char* js_visp_norm( char* oldtext, pool pp )
{
    char*   newtext;
    int     first, last, i, j;

    for ( first = 0;  first < strlen(oldtext);  first++ )
    {
        if ( oldtext[first] != ' '  &&
             oldtext[first] != '\t' &&
             oldtext[first] != '\n' ) break;
    }

    for ( last = strlen(oldtext)-1;  last >= 0;  last-- )
    {
        if ( oldtext[last] != ' '  &&
             oldtext[last] != '\t' &&
             oldtext[last] != '\n' ) break;
    }

    if ( ( newtext = pmalloc( pp, last-first+2 ) ) != NULL )
    {
        for ( i = first,  j = 0;
              i <= last;
              i++, j++ )
        {
            newtext[j] = tolower( oldtext[i] );
        }

        newtext[last-first+1] = '\0';
    }

    return newtext;
}
#undef FUNCNAME


/*************************************************************************
 *  js_visplist_exist -- Checks whether VISP Search Base lookup table
 *                       has been read into memory.
 *
 *  parameters
 *      None.
 *
 *  returns
 *      1 if there is a valid lookup table (even if the load failed).
 *      0 if there is no lookup table available.
 */
int js_visplist_exist( )
{
    return ( js__visplist != NULL );
}


/*************************************************************************
 *  js_visplist_load -- load the VISP Search Base lookup table file.
 *
 *  Reads an XML Search Base lookup table file and stores a pointer
 *  to the root node in the global variable "js_vsbase".
 *  NOTE:  This is always called at cold start, but may also be called
 *  later to re-read the file in response to an HUP signal.
 *
 *  parameters
 *      cw_xdb_node -- Pointer to the configuration file's <xdb...>
 *                     section that loads the C&W LDAP shared object
 *                     library (including this code).
 *                     (NOTE: "cw_xdb_node" is currently unused here.)
 *      file        -- Pointer to character string containing the full
 *                     filename of the Search Base lookup table file
 *                     to read.
 *
 *  returns
 *      1 if there is a valid lookup table (even if the load failed).
 *      0 if there is no lookup table available.
 */
#define FUNCNAME    "js_visplist_load"

int js_visplist_load( xmlnode cw_xdb_node, char* file )
{
    /* VISPLISTXML is the default name for the Search Base lookup
       table file - to be defined by the build system */

    char    def[]       = VISPLISTXML;
    char*   realfile    = (char*)def;
    xmlnode oldvisplist = js__visplist;
    int     file_reload = ( oldvisplist != NULL );

    /* If no file name is specified, use the default file path */

    if( file != NULL )  realfile = file;

    /* Read and parse the VISP Search Base lookup table file */

    if ( ( js__visplist = xmlnode_file( realfile ) ) == NULL )
    {
        /* Read or parse failed.  Do we have an existing lookup */
        /* table to fall back to?                               */

        if( file_reload )
        {
            /* This was a failed HUP-triggered reload -- */
            /* -- so restore the previous lookup table.  */

            js__visplist= oldvisplist ;
            log_error( ZONE, ": " FUNCNAME "(): "
                       "VISP lookup table update using %s failed, "
                       "reverting to existing lookup", realfile );
        }
        else
        {
            /* No fallback lookup table - report loading failure */

            log_error( ZONE, ": " FUNCNAME "(): "
                       "VISP lookup table update using %s failed",
                       realfile );
        }
    }

    /* If a successful reload, take any required concomitant actions. */

    if( file_reload )
    {
        /*
         *  FIXME: Notify modules, services etc to use the new table.
         *  FIXME: Free the old table.
         */
    }

    return ( js__visplist != NULL );
}
#undef FUNCNAME


/*************************************************************************
 *  js_vispnode -- Get a node in the VISP Search Base Look-Up Table.
 *
 *  Parameters:
 *      query -- Pointer to a string containing the path through the
 *               tag hierarchy to the desired tag.
 *               eg. for the file:
 *                   <fname><foo><bar>bar value</bar><baz/></foo></fname>
 *               use "foo/bar" to retreive the bar node
 *
 *  Returns:
 *      The root visplist node if the pointer to the query is NULL,
 *      or a pointer to the xmlnode specified in the query,
 *      or NULL if the query fails (i.e. node is not found).
 */
#define FUNCNAME    "js_vispnode"

xmlnode js_vispnode( char* query )
{
    xmlnode retpnode;
    log_debug( ZONE, ": " FUNCNAME "(): "
                     "VISPlist query %s - enter", query );

    if ( js__visplist == NULL )        /* Has table been loaded? */
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                   "VISP lookup attempt before table has been loaded" );
        return NULL;
    }

    /* was a query specified? */
    if( query == NULL )
        return js__visplist;      /* return the root config node */
    else
    {
        retpnode = xmlnode_get_tag( js__visplist, query );

        log_debug( ZONE, ": " FUNCNAME "(): "
                         "VISPlist query %s - leave (ret=0x%08x)",
                         query, retpnode );

        if ( retpnode == NULL )
            log_debug( ZONE, ": " FUNCNAME "(): "
                             "Failed to locate <%s> in table",
                             query );

        return retpnode;          /* return the specified tag (node) */
    }
}
#undef FUNCNAME


/*************************************************************************
 *  js_vispfind -- Locate a VISP entry, given a tag name and expected
 *                 content, and return that VISP's node.
 *
 *  This function can be used to find the node for a particular VISP,
 *  given either its long or short name.
 *
 *  Parameters:
 *      tag   -- Pointer to a string containing the name of the tag to
 *               be used in the search (e.g. "name" or "full").
 *      data  -- Pointer to a string containing the data content that
 *               is to be matched in the specified tag.
 *
 *  Returns:
 *      Pointer to the <visp> node for the specified VISP,
 *      or NULL if the query fails (i.e. the data is not matched).
 */
#define FUNCNAME    "js_vispfind"

xmlnode js_vispfind( char* vtag, char* vdata )
{
    xmlnode xnvisp;
    char*   ndata;         /* Normalized version of "vdata" */
    char*   nname;         /* Name from <visp> element */
    char*   ddata;         /* Data from specified tag */

    log_debug( ZONE, ": " FUNCNAME "(): "
                     "VISP search for \"<%s>%s\" - enter",
                     vtag, vdata );

    /* First time only - locate the <vbases> node in the lookup table.
     * (This will also check that the lookup table has been loaded and
     *report errors.)
     */
    if ( js__vispbases == NULL )
    {
        if ( ( js__vispbases = js_vispnode( CW_VBASES ) ) == NULL )
            return NULL;

        js__vispfirst = xmlnode_get_firstchild( js__vispbases );
    }

    /* Does the <vbases> node have children (i.e. <visp> entries)? */

    if ( ( xnvisp = js__vispfirst ) == NULL )
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                         "<" CW_VBASES "> has no children" );
        return NULL;
    }

    if( vtag == NULL )        /* Must specify a tag name */
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                         "Tag name not specified in call" );
        return NULL;
    }

    if( vdata == NULL )       /* Must specify the data to match */
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                         "Data content not specified in call" );
        return NULL;
    }

    /* Normalize the supplied string (once only). */

    ndata = js_visp_norm( vdata, xnvisp->p );

    /* Traverse the <vbases> node's children looking for all <visp>
     * elements.  (Any other element type is ignored.)
     */
    for ( ;
          xnvisp != NULL;
          xnvisp = xmlnode_get_nextsibling( xnvisp ) )
    {
        /* Get the name of each child node.  We need to check
         * and discard <vbase> children that have no name.
         * These 'empty' nodes seem to occur as a result of
         * newlines between the <vbases> and <visp> tags in
         * the XML document.
         */
        nname = xmlnode_get_name( xnvisp );

        if ( nname != NULL )
        if ( strcmp( CW_VISP, nname ) == 0 )
        {
            /* We have a <visp> entry - so find the specified tag
             * within it and see if it matches the specified data.
             */
            ddata = xmlnode_get_tag_data( xnvisp, vtag );

            if ( strcmp( js_visp_norm( ddata, xnvisp->p ), ndata )
                 == 0 )
                break;
        }
    }

    if ( xnvisp == NULL )
        log_debug( ZONE, ": " FUNCNAME "(): "
                         "Could not find \"%s\" VISP", vdata );

    return xnvisp;
}
#undef FUNCNAME


/*************************************************************************
 *  js_visp_s_to_l -- Given a short VISP name, return the long
 *                    version (as stored in the <full> element).
 *
 *  Parameters:
 *      sname -- Pointer to a string containing the VISP's short name.
 *
 *  Returns:
 *      Pointer to the Base string for the specified VISP,
 *      or NULL if the query fails (i.e. VISP is not found).
 */
#define FUNCNAME    "js_visp_s_to_l"

char* js_visp_s_to_l( char* sname )
{
    xmlnode xnvisp;

    log_debug( ZONE, ": " FUNCNAME "(): "
                     "VISP short-to-long for \"%s\" - enter",
                     sname );

    /* In the lookup table, locate the <visp> node that has the
     * specified short name.  ("js_vispfind" will also check that
     * the lookup table has been loaded and report any errors.)
     */
    if ( ( xnvisp = js_vispfind( CW_VNAME, sname ) ) == NULL )
        return NULL;          /* Abandon if not found */

    /* We have found the node for the requested VISP */

    return xmlnode_get_tag_data( xnvisp, CW_VFULL );
}
#undef FUNCNAME


/*************************************************************************
 *  js_visp_l_to_s -- Given a long VISP name, return the short
 *                    version (as stored in the <name> element).
 *
 *  Parameters:
 *      lname -- Pointer to a string containing the VISP's long name.
 *
 *  Returns:
 *      Pointer to the Base string for the specified VISP,
 *      or NULL if the query fails (i.e. VISP is not found).
 */
#define FUNCNAME    "js_visp_l_to_s"

char* js_visp_l_to_s( char* lname )
{
    xmlnode xnvisp;

    log_debug( ZONE, ": " FUNCNAME "(): "
                     "VISP long-to-short for \"%s\" - enter",
                     lname );

    /* In the lookup table, locate the <visp> node that has the
     * specified full name.  ("js_vispfind" will also check that
     * the lookup table has been loaded and report any errors.)
     */
    if ( ( xnvisp = js_vispfind( CW_VFULL, lname ) ) == NULL )
        return NULL;          /* Abandon if not found */

    /* We have found the node for the requested VISP */

    return xmlnode_get_tag_data( xnvisp, CW_VNAME );
}
#undef FUNCNAME


/*************************************************************************
 *  js_vispbase -- Given a VISP name, return its Base string.
 *
 *  Parameters:
 *      vname -- Pointer to a string containing the short name of the
 *               VISP (as stored in the <name> element).
 *
 *  Returns:
 *      Pointer to the Base string for the specified VISP,
 *      or NULL if the query fails (i.e. VISP is not found).
 */
#define FUNCNAME    "js_vispbase"

char* js_vispbase( char* vname )
{
    xmlnode xnvisp;

    log_debug( ZONE, ": " FUNCNAME "(): "
                     "VISP lookup for \"%s\" - enter", vname );

    /* In the lookup table, locate the <visp> node that has the
     * specified full name.  ("js_vispfind" will also check that
     * the lookup table has been loaded and report any errors.)
     */
    if ( ( xnvisp = js_vispfind( CW_VFULL, vname ) ) == NULL )
        return NULL;          /* Abandon if not found */

    /* We have found the node for the requested VISP */

    return xmlnode_get_tag_data( xnvisp, CW_VBASE );
}
#undef FUNCNAME

