/* Cable and Wireless Customization of Jabber Instant Messaging Server */

/* O.D.G. Griffiths     Started: 28-Dec-00 */

/* ATTENTION: Jabber Server Version: 1.4 */

/* Source file name:    cw_thrdm.c */

/* --------------------------------------------------------------------------
 *
 * License
 *
 * The contents of this file are subject to the Jabber Open Source License
 * Version 1.0 (the "License").  You may not copy or use this file, in either
 * source code or executable form, except in compliance with the License.  You
 * may obtain a copy of the License at http://www.jabber.com/license/ or at
 * http://www.opensource.org/.  
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.  See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyrights
 * 
 * Portions created by or assigned to Jabber.com, Inc. are 
 * Copyright (c) 1999-2000 Jabber.com, Inc.  All Rights Reserved.  Contact
 * information for Jabber.com, Inc. is available at http://www.jabber.com/.
 *
 * Portions Copyright (c) 1998-1999 Jeremie Miller.
 * 
 * Acknowledgements
 * 
 * Special thanks to the Jabber Open Source Contributors for their
 * suggestions and support of Jabber.
 * 
 * --------------------------------------------------------------------------*/


/***********************************************/
/* Thread Pool Manager for LDAP Access Threads */
/***********************************************/
/*
 * Phase 2 of the C&W Jabber modifications requires greater scalability
 * than Phase 1.  However, LDAP access is a relatively slow process
 * because it involves multiple exchanges with remote computers.
 * Execution of the authentication process is blocked while a query to the
 * LDAP Directory take place.  If only a single thread were used for LDAP
 * enquiries, authentication requests would be queued one behind the other,
 * severely limiting the rate at which they could be processed. To avoid
 * this queueing, Phase 2 processes multiple LDAP queries in parallel.  To
 * proceed in parallel, each query must be on a separate execution thread.
 * The multiple threads are managed by this Thread Pool Manager.
 *
 * Creation and destruction of threads constitutes an overhead, so, rather
 * than creating a thread every time the LDAP directory server is queried,
 * a pool of threads is maintained, ready for use.
 *
 * A relatively small fixed number of threads is pre-created (the number
 * is defined by LMT_THREADS), each with its own PTH message port.  An
 * extra 'master overflow' message port is created to handle all requests
 * that arrive while all the threads are busy.  Requests on the overflow
 * port thus drop to the efficiency of a single-thread solution.
 *
 * The following structures and functions have been adapted from the "mtq"
 * Managed Thread Queue Equivalents in the "mtq.c" source file (formerly in
 * "base_load.c"). To differentiate this code from the "mtq" originals,
 * "mtq" has been replaced by "lmt" and "mth" has been replaced by "lth"
 * throughout this module and also in "cw_jabberd.h".
 * "lmt" stands for LDAP Managed Thread.
 *
 * A major difference between this source module and "mtq.c" is that
 * "mtq.c" provides two modes of operation, whereas this module does not.
 * "mtq.c" not only handles single transactions, but can also handle
 * ongoing streams of traffic with multiple messages being fed to the same
 * thread.  In the "xdb_ldap" context, there is no requirement to keep an
 * ongoing channel open.  Each invocation is like a single database query,
 * say, yielding a particular result.  So this module is simpler than
 * "mtq.c".
 *
 *
 * OUTSTANDING ISSUES
 * ==================
 * 1)  ##########
 *     ##########
 */


#include "cw_sysdef.h"
#include "cw_sbase.h"


/*******************/
/* Local Constants */
/*******************/


/**************************************************/
/* Control Block Structures for Thread Management */
/**************************************************/

/*** Control Block passed to Thread at Start-Up ***/

typedef struct lth_struct
{
    pth_msgport_t  mp;    /* Message port through which the thread */
                          /* will receive its assignments          */
    pth_t          id;
    pth_event_t    mpevt; /* Event ring for waiting on message port */
    pool           p;
    int            busy;
    LDAP*          pld;   /* Pointer to LDAP connection handle */
} *lth, _lth;


/*** LMT Global Data Block  (only ever one instance of this) ***/

typedef struct lmtglobal_struct
{
    lth            all[LMT_THREADS];  /* Array of pointers to all the */
                                      /* thread control blocks        */
    pth_msgport_t  mp;           /* Master port to queue overflow if  */
                                 /* all threads are currently in use. */
    int            overflow;     /* Number of requests currently on */
                                 /* the overflow queue port         */
    mod_ldap_conf  cfg;          /* Pointer to LDAP Config details */
                                 /* (needed to open one-per-thread */
                                 /* LDAP connections)              */
} *lmtglobal, _lmtglobal;


/*** Control Block passed to Thread for Each Transaction ***/

/* This structure contains all the information passed to the main-line
 * function shared by all threads, to initiate a transaction on a
 * thread.  One is constructed for every transaction that is initiated.
 */
typedef struct lmtcall_struct
{
    pth_message_t  head;  /* The standard PTH message header */
    lmt_callback   f;     /* Function to run within the thread */
    void*          arg;   /* The data to pass to the callback function */
} _lmtcall, *lmtcall;


/***************/
/* Global Data */
/***************/

lmtglobal  lmt__global = NULL;  /* Pointer to one and only instance of */
                                /* the global data block (see above)   */

/****************************/
/* Local Function Templates */
/****************************/

void *lmt_main( void *arg );


/*******************************************************************************
 * ldap_threadman_init -- Initialization for the LDAP thread pool.
 *
 * Once-only cold-start initialization.  Constructs the global data block,
 * including the (fixed-size) thread pool.
 *
 * Parameters:
 *      None.
 */
#define FUNCNAME    "ldap_threadman_init"

void ldap_threadman_init( mod_ldap_conf cfg )
{
    int         n;
    pool        newp;
    lth         tdata = NULL;
    pth_attr_t  attr;

    if( lmt__global != NULL )
    {
        log_error( ZONE, ": " FUNCNAME "(): "
                         "LDAP thread pool initialized more than once." );
        return;
    }

    /* Allocate space for global data -- once only, with a real (non-pool)
     * malloc -- and create the 'overflow' port that will be used to handle
     * any requests that arrive when all threads are busy.
     */
    lmt__global      = malloc( sizeof(_lmtglobal) );
    lmt__global->mp  = pth_msgport_create( "lmt_overflow" );
    lmt__global->cfg = cfg;

    /* Create the pool of threads.  (In the global data, set up the array
     * of pointers to the control blocks for each individual thread.)
     */
    for( n=0; n<LMT_THREADS; n++ )
    {
        /* Each thread has its own memory pool and message input port (as
         * well as stack, stack pointer and program counter).  Each thread
         * starts running the code of "lmt_main" (below) as soon as it gets
         * an opportunity to run (i.e. reaches the top of the Scheduler's
         * "Ready" queue).
         */
        newp       = pool_new();
        tdata      = pmalloco( newp, sizeof(_lth) );
        tdata->p   = newp;
        tdata->mp  = pth_msgport_create( "lth" );
        tdata->pld = NULL;
        attr  = pth_attr_new();
        pth_attr_set( attr, PTH_ATTR_PRIO, PTH_PRIO_MAX );

        if ( ( tdata->id = pth_spawn( attr, lmt_main, (void *)tdata ) )
             == NULL )
        {
            log_error( ZONE, ": " FUNCNAME "(): "
                             "Failed to spawn LDAP pool thread." );
        }

        pth_attr_destroy( attr );
        lmt__global->all[n] = tdata;   /* Store pointer to each thread  */
                                       /* control block in global table */
    }
}
#undef FUNCNAME


/*******************************************************************************
 * ldap_threadman_send -- Initiate handling of an LDAP access transaction,
 *                        on its own thread.
 *
 * This function launches a transaction that is to be executed on its own
 * thread (in parallel with other LDAP access transactions).  If a thread
 * is available, the transaction is assigned to that thread immediately.
 * If all threads are currently busy, the transaction is added to a holding
 * queue, from which transactions are activated as threads become free.
 * (Activation of transactions from the overflow queue is handled in the
 * "lmt_main" main-line function that is used by all the threads.  This
 * function merely adds transactions to the queue.)
 *
 * Different transactions require different LDAP access actions to be
 * taken, so these are passed to "lmt_send" in the form of a call-back
 * function, which is then invoked from the thread's "lmt_main" main-line
 * function.  The pointer to the call-back function is passed to "lmt_main"
 * in the argument bundle defined in the "lmtcall" structure.
 *
 * Parameters:
 *      p   -- (pool)   Pointer to the pool to be used.
 *      f   -- (lmt_callback)  Pointer to the callback function that will
 *                      carry out the required actions for the transaction
 *                      that is being scheduled here.
 *      arg -- (void*)  Pointer to the parameter block for this thread.
 */
#define FUNCNAME    "ldap_threadman_send"

void ldap_threadman_send( pool p, lmt_callback f, void *arg )
{
    lmtcall        c;
    int            n;
    pth_msgport_t  mp = NULL;    /* Destination port for msg */


    /* Find a waiting thread, unless all are busy. */

    for ( n = 0;  n < LMT_THREADS;  n++ )
    {
        if ( lmt__global->all[n]->busy == 0 )
        {
            /* We have a free thread, so select its entry port and
             * flag that thread as busy.
             */
            mp = lmt__global->all[n]->mp;
            lmt__global->all[n]->busy = 1;
            break;
        }
    }

    /* If we didn't find an available thread, set the 'overflow' message
     * port as its destination, to get picked up when a thread becomes free.
     */
    if ( mp == NULL )
    {
        log_debug( ZONE, ": " FUNCNAME "(): %d overflowing %X",
                            lmt__global->overflow, arg );
        mp = lmt__global->mp;
        lmt__global->overflow++;

        /* The original code in "mtq.c" contained the following comment:
         *
         *   "XXX this is a race condition in pthreads.. if the overflow
         *   is not put on the mp, before a worker thread checks the mp
         *   for messages, then it will set this variable to 0 and not
         *   check the overflow mp until another item overflows it."
         *
         * GG: Assume that "pthreads" here refers to Posix threads, the
         *     difference being that Posix uses preemptive multithreading
         * whereas PTH uses cooperative multithreading, so Posix can break
         * in between any two machine instructions but PTH has to wait for
         * a call to a suitable PTH API function for the scheduler to take
         * control.
         *
         * I have avoided this problem here by simply checking the overflow
         * queue port every time (in "lmt_main()"), so any race condition
         * in resetting this counter is irrelevant.  (It is now used only
         * for debug display.)
         */
    }

    /* Construct control block for this call (in the supplied pool) and
     * send the request to the chosen thread, or overflow queue.
     */
    c      = pmalloco( p, sizeof(_lmtcall) );
    c->f   = f;
    c->arg = arg;
    pth_msgport_put( mp, (pth_message_t *)c );
    return;
}
#undef FUNCNAME


/*******************************************************************************
 * lmt_main-- Main-line function for any of the pool threads.
 *
 * This first picks up the arguments passed to it, which include the
 * message port that has been assigned to accept 'input' to the current
 * thread.  (Each thread has its own such 'input' port, assigned by
 * "ldap_threadman_init()".)
 * The main-line then enters a do-forever loop, which first checks the
 * overflow queue port to see if it has any requests queued.  If not, it
 * checks the thread's own input port to see if a request has been sent
 * to it.
 *
 * Parameters:
 *      arg -- (void*)  Pointer to the "lth" parameter block for this thread.
 */
#define FUNCNAME    "lmt_main"

void *lmt_main( void *arg )
{
    lth          t = (lth)arg;
    lmtcall      c;
    result       retval;

    log_debug( ZONE, ": " FUNCNAME "(): %X thread starting",t->id );

    /* create an event ring for receiving messges */

    t->mpevt = pth_event( PTH_EVENT_MSG, t->mp );

    while( 1 )       /* Loop forever */
    {
        /* Before checking this thread's own 'inout' port, see if there
         * are any requests already waiting on the overflow queue.
         * The equivalent code in "mtq.c" uses a separate flag
         * ("mtq__master->overflow") to indicate when something is waiting
         * on the overflow queue.  However, as a comment in "mtq.c" says,
         * this raises a potential race condition if and when PTH is
         * replaced by Posix pthreads, because these use preemptive
         * multithreading (and can therefore potentially break in between
         * the call to "pth_msgport_get" and the subsequent instruction to
         * clear "mtq__master->overflow" if the queue proves to be empty.
         *
         * In fact, in "mtq.c", the "mtq__master->overflow" flag only
         * avoids the execution overhead of an unnecessary call to
         * "pth_msgport_get".  But in the context of LDAP queries, such an
         * overhead is negligible compared with the execution time of the
         * LDAP query itself, and in any case, LDAP queries are relatively
         * rare events, compared with the session traffic handled by
         * "mtq.c".
         *
         * So here we take the simplistic approach - we simply interrogate
         * the overflow queue itself every time.   That way, we will never
         * miss any waiting overflow, even when the code is converted to
         * pthreads.  (The overflow counter is now only used for debug
         * display purposes.)
         */

        if ( ( c = (lmtcall)pth_msgport_get(lmt__global->mp) ) != NULL )
        {
            /* We got a request off the overflow queue, so use that. */
        }
        else     /* Otherwise... no overflow, so wait on our own port. */
        {
            /* Clear the overflow count (only used for debug display). */
            lmt__global->overflow = 0;

            log_debug( ZONE, ": " FUNCNAME "(): "
                             "%X thread waiting at pth_wait", t->id );
            t->busy = 0;     /* Indicate that this thread is free */


            /*** Wait for a message on this thread's own input port ***/
            /*   (This is where we relinquish and regain control)     */

            pth_wait( t->mpevt );      /* Sleep until something happens */

            /* NOTE:  The "t->busy = 1" command below is as per the
             * ====   original "mtq.c" code.  Most of the time, this
             *        is redundant, as it is already set in the
             * "ldap_threadman_send()" function, above, when the thread
             * is allocated, which is the safe thing to do (as the
             * "ldap_threadman_send()" function could get called several
             * times before the thread actually gets to run).
             * However, I assume that the line below is necessary for the
             * case where we have first picked up one or more requests
             * from the overflow queue.  In this case, we will execute
             * the "t->busy = 0" above and go into the wait, to find that
             * the non-overflow transaction that was assigned to this
             * thread is still sitting there, so we will come straight
             * out of the pth_wait again (but with the "busy" flag now
             * zero unless we set it again, below).
             * In this case there could cause a race condition between
             * this function and "ldap_threadman_send()", because if a
             * context switch takes place when this thread enters the
             * "pth_wait()" call above, "ldap_threadman_send()", might
             * get control, see the the "busy" flag for this thread is
             * zero and send it another transaction.  However, this
             * should be benign, because this thread will then simply
             * process both transactions one after the other (and any
             * more that get added to the queue in the meantime, until
             * the queue really is empty).              ODGG 14-Feb-01
             */
            t->busy = 1;       /* Indicate that this thread is now busy */

            log_debug( ZONE, ": " FUNCNAME "(): "
                             "Re-entering %X thread from pth_wait", t->id );

            c = (lmtcall)pth_msgport_get( t->mp );    /* Get the message */
            if( c == NULL ) continue;     /* Bail out if failed to get it */
        }

        /* Connection to the LDAP Server is a two-stage process.  Before
         * search functions such as "ldap_search()" can be used, connection
         * must first be established by calling "ldap_init()" followed by a
         * bind operation such as "ldap_simple_bind_s()".  (In the C&W
         * context, the bind also validates the Jabber server's authority to
         * access the X.500 directory.)
         * The original CW Phase 1 implementation opened and closed the LDAP
         * connection for every user authentication (each one thus requiring
         * two accesses to the remote X.500 directory).  Here, we try to be
         * more efficient by keeping the LDAP connection open once it is
         * established and reusing it for each successive LDAP search, thus
         * avoiding the preliminary bind on each.
         * To avoid any issues of concurrent use of an LDAP connection,
         * each thread maintains its own connection.
         * Since the Jabber server is trying to maintain 100% uptime, it
         * is possible that access to the X.500 directory may be lost from
         * time to time, causing an LDAP transaction to fail.  If this
         * happens, the LDAP connection is closed, so that we can start
         * again from scratch the next time around.
         */
         if ( t->pld == NULL )
         {
            /* We don't currently have the LDAP connection set up, so
             * try to establish one.  ("js_cw_ldap_init()" performs both the
             * "ldap_init()" and "ldap_simple_bind_s()" calls.)
             */
            if ( ( t->pld = js_cw_ldap_init( lmt__global->cfg ) ) == NULL )
            {
                /* We failed to connect. */

                log_error( ZONE, ": " FUNCNAME "(): "
                           "Initial connection to LDAP server failed." );
            }
        }

        /* The callback function is responsible for sending off the response
         * to the query (via the "deliver()" function), whether it succeeds
         * or fails.  We must therefore always call the callback function,
         * irrespective of whether we succeeded or failed here in opening
         * the LDAP connection.  The callback function detects such failure
         * by checking whether the passed LDAP handle is NULL.
         */
        log_debug( ZONE, ": " FUNCNAME "(): "
                         "%X invoking LDAP action %X", t->id, c->arg );

        retval = (*(c->f))( c->arg, t->pld );


        /* If there was any kind of failure, close the LDAP connection, so
         * that we try to re-open it afresh next time around.
         */
        if ( retval != r_DONE )
        {
            /* Release connection to the LDAP server. */

            js_cw_ldap_drop( lmt__global->cfg, t->pld );
            t->pld = NULL;
        }
    }

    /* ##########  EXECUTION WILL NEVER REACH THE CODE BELOW  ########## */
    /* ##########  AS THERE ARE NO EXITS FROM THE LOOP ABOVE  ########## */
    /* ##########  (BUT THIS IS HOW IT WAS IN ORIGINAL mtq.c) ########## */

    /* free all memory stuff associated with the thread */

    pth_event_free(t->mpevt,PTH_FREE_ALL);
    pth_msgport_destroy(t->mp);
    pool_free(t->p);
    return NULL;
}
#undef FUNCNAME


/*******************************************************************************
 * ldap_threadman_shutdown -- Clean-up function at program shutdown.
 *
 * Closes all open threads, LDAP connections, etc.
 *
 * Parameters:
 *      arg -- (void*)  Pointer to the CW LDAP handler's global parameter
 *                      block.
 */
#define FUNCNAME    "ldap_threadman_shutdown"

void ldap_threadman_shutdown( void *arg )
{
    int     n;      /* Loop counter */
    lth     t;      /* Pointer to thread control block */

    log_debug( ZONE, ": " FUNCNAME "(): CW LDAP Query Handler SHUTDOWN" );

    /* Work through all the threads. */

    for ( n = 0;  n < LMT_THREADS;  n++ )
    {
        t = lmt__global->all[n];

        if ( t->busy != 0 )
        {
            /* In the future, might need to do something if threads are
             * still busy when we are trying to shut them down.  But at
             * the moment we are only using this function to shut down
             * the LDAP connections and thread-busy is not a show-
             * stopper for this.
             */
        }

        /* If this thread's LDAP connection is open, close it. */

        if ( t->pld != NULL )
        {
            /* Release connection to the LDAP server. */

            js_cw_ldap_drop( lmt__global->cfg, t->pld );
            t->pld = NULL;
        }

        /*** Free all memory stuff associated with the thread. ***/

        /* Including the following "pth_event_free()" call caused the
         * application to dump core later in the shutdown process (AFTER
         * exiting from this function), so the following lines have been
         * commented out).  Thread closure has also been omitted, as the
         * thread should only be closed after its facilities have been
         * cleaned up.  We therefore rely on the operating system to
         * clear up everything when the Jabber server's process is
         * terminated.
         */
        /* pth_event_free( t->mpevt, PTH_FREE_ALL ); */
        /* pth_msgport_destroy( t->mp );             */
        /* pool_free( t->p );                        */
    }
}
#undef FUNCNAME


