/* Cable and Wireless Jabber modification code - for vcard access */

#include "cw_sysdef.h"


/***************/
/* Global Data */
/***************/

/*** Function Templates (for forward references to local functions) ***/

int js_cw_event_mon( void* arg );
int js_cw_waitx( int rc, LDAPMessage** result, waitevt w );
result cw_vcard_access (dpacket p, LDAP* ld, mod_ldap_conf cfg);
xmlnode cw_ldap_vcard_get(dpacket p, LDAP* ld, mod_ldap_conf cfg, int size);
result cw_ldap_vcard_set(dpacket p, LDAP* ld, mod_ldap_conf cfg, xmlnode data, int size);
int cw_find_vcard_size(mod_ldap_conf cfg);
xmlnode cw_build_vcard_data(LDAP* ld,LDAPMessage* e,mod_ldap_conf cfg,xmlnode data);
LDAPMod** cw_strip_vcard_data (mod_ldap_conf cfg, xmlnode data, int size);
int cw_check_n (const char* name);
int cw_check_adr (const char* name);
int cw_check_org (const char* name);
int cw_check_tel (const char* name);
int cw_check_email (const char* name);
xmlnode cw_get_adr_node(xmlnode vcard,int type);
xmlnode cw_get_email_node(xmlnode vcard,int type);
xmlnode cw_get_tel_node(xmlnode vcard,int type, const char* name);
xmlnode cw_insert_n_node (xmlnode data, char** vals, char* name);
xmlnode cw_insert_org_node (xmlnode data, char** vals, char* name);
xmlnode cw_insert_adr_node(xmlnode data,char** vals,char* name,int adr);
xmlnode cw_insert_tel_node(xmlnode data,char** vals,char* name,int tel);
xmlnode cw_insert_email_node(xmlnode data,char** vals,char* name,int email);
char* cw_tokenise_suffix(char* name);

/*******************/
/* Local Constants */
/*******************/

#define NONE 0
#define WORK 1
#define HOME 2
#define PREF 1
#define OTHER 2

/* this function is used to access the LDAP directory and set or get vcard information */

/* parameters - dpacket p - dpacket to the session
 *              LDAP* ld - ldap connection configuration info
 *              mod_ldap_conf config - configuration info for the directory/schema
 *         	HASHTABLE cache - cache for LDAP roster
 *
 * result - returns success or not
 */

result cw_vcard_access (dpacket p, LDAP* ld, mod_ldap_conf cfg)
{
  xmlnode data=NULL;
  result ret=r_ERR;
  int set=0;
  
  log_debug(ZONE,"handling LDAP request %s",xmlnode2str(p->x));
  
  if(j_strcmp(xmlnode_get_attrib(p->x,"type"), "set") == 0)
    {
      /* vcard modify access */
      ret=cw_ldap_vcard_set(p,ld,cfg,p->x,cw_find_vcard_size(cfg));
      set=1;
    }
  
  else
    {
      /* vcard read access only */
      data=cw_ldap_vcard_get(p,ld,cfg,cw_find_vcard_size(cfg));
    }
  
  if(data != NULL)
    { /* cool, send them back a copy of the data */
      xmlnode_hide_attrib(xmlnode_insert_tag_node(p->x, data),"xdbns");
      ret=r_DONE;
    }

  /* check this */  
  if(ret!=r_ERR && !set)
    {
      xmlnode_put_attrib(p->x,"type","result");
      xmlnode_put_attrib(p->x,"to",xmlnode_get_attrib(p->x,"from"));
      xmlnode_put_attrib(p->x,"from",jid_full(p->id));
      deliver(dpacket_new(p->x), NULL); /* dpacket_new() shouldn't ever return NULL */
    }
  return ret;      
}




/* this function is to get roster information from the directory - 
 * it is modeled on the authentication function */

#define FUNCNAME "cw_ldap_vcard_get"

xmlnode cw_ldap_vcard_get(dpacket p, LDAP* ld, mod_ldap_conf cfg, int size)
{
  waitevt w;
  pth_event_t evt;
  int have_evt=FALSE,i=0;
  spool sp;
  char *filter, *attribs[size], *dn=NULL, *error;
  LDAPMessage* qresult;
  LDAPMessage* e=NULL;
  xmlnode data=NULL;
  vcard q;
  
    log_debug(ZONE,": " FUNCNAME " (): Started LDAP vcard retrieval for user");
  
  /* checking for connection to LDAP server */
  
  if (ld==NULL)
    {
      /* no connection, return NULL data */
      log_error(ZONE,": " FUNCNAME "(): No connection to LDAP server");
      return data;
    }
  
  /* create data object to be sent to wait-event */
  
  w=pmalloc(p->id->p,sizeof(_waitevt));
  w->ld=ld;
  
  /* construct LDAP binding argument and define what attributes are to 
     be returned from the search */
  
  q=cfg->vcard_attrib;
  
  for (i=0;i<size;i++)
    {
      if (q)
	{
	  attribs[i]=q->mapping;
	  q=q->next;
	}
    }
  /* null terminate last attribue */
  attribs[size-1]=NULL;

  sp=spool_new(p->id->p);
  spooler(sp,cfg->user_attrib,"=",p->id->user,sp);
  filter=spool_print(sp);
  
  log_debug(ZONE,": " FUNCNAME "(): filter: %s",filter);
  
  /* address query to the LDAP directory */
  
  if ((w->msgid=ldap_search(w->ld,cfg->base,LDAP_SCOPE_SUBTREE,
			    filter,attribs,0))
      ==-1)
    {
      /* error in the search call */
      error=ldap_err2string(ldap_result2error(ld,w->result,1));
      log_error(ZONE,": " FUNCNAME "(): LDAP result \"%s\" occured during vcard retrieve operation",error);
      return data;
    }
  
  /* wait for response to come back from LDAP directory */
  evt = pth_event(PTH_EVENT_FUNC,&js_cw_event_mon,(void *)w,
		  pth_time(0,500000));
  pth_wait(evt); /* block until LDAP responds */
  have_evt=TRUE;
  
  if (w->result==NULL)
    {
      /* an error occured during wait */
      log_error(ZONE,": " FUNCNAME "(): Error occured while waiting for result");
      pth_event_free(evt,PTH_FREE_ALL);
      return data;
    }
  
  log_debug(ZONE,": " FUNCNAME "(): result: 0x%08x", w->result);
  
  /* check the result of the search */

  if (w->rc!= LDAP_RES_SEARCH_ENTRY)
    {
      /* add extra checking here to distinguish the LDAP error */
      error=ldap_err2string(ldap_result2error(ld,w->result,1));
      log_error(ZONE,": " FUNCNAME "(): LDAP result \"%s\" occured during roster retrieve operation",error);
    }
  
  else
    {
      log_debug(ZONE,": " FUNCNAME "(): LDAP result should be LDAP_RES_ %d",LDAP_SUCCESS);
      qresult=w->result;
      
      if ((e=ldap_first_entry(w->ld,qresult))==NULL)
	{
	  log_error(ZONE,": " FUNCNAME "(): No first entry from LDAP result");
	}
      else if ((dn=ldap_get_dn(w->ld,e))==NULL)
	{
	  log_error(ZONE,": " FUNCNAME "(): No dn from first entry in LDAP result");
	}
      else
	{
	  log_debug(ZONE,": " FUNCNAME "(): dn is %s",dn);

	  /* check here to see if you need to pass in the dn and the ber */
	  
	  /* build the vcard xml data */
	  if ((data=xmlnode_new_tag("vCard"))!=NULL)
	    {
	      xmlnode_put_attrib(data,"version","3.0");
	      xmlnode_put_attrib(data,"xmlns","vcard-temp");
	      data=cw_build_vcard_data(w->ld,e,cfg,data);
	    }
	}
    }

  if (e!=NULL)
    free(e);
  if (dn!=NULL)
    free(dn);
  
  if (have_evt)
    pth_event_free(evt,PTH_FREE_ALL);
  
  log_debug(ZONE,": " FUNCNAME "():LDAP roster access completed");
  
  return data;
}
#undef FUNCNAME




/* this function is used to write data to the ldap directory */

#define FUNCNAME "cw_ldap_vcard_set"

result cw_ldap_vcard_set (dpacket p, LDAP* ld, mod_ldap_conf cfg, xmlnode data, int size)
{
  waitevt w;
  pth_event_t evt;
  int have_evt=FALSE;
  spool sp;
  char *filter,*attribs[size][2], *error;
  LDAPMod modified_data[size],*modified[size];
  result retval=r_ERR;
  vcard q;
  int i=0,adr=NONE;
  xmlnode vcard=xmlnode_get_firstchild(data);
  xmlnode temp;


  log_debug(ZONE,": " FUNCNAME " (): Started LDAP vcard modify for user");
  
  /* checking for connection to LDAP server */

  if (ld==NULL)
    {
      /* no connection, return NULL data */
      log_error(ZONE,": " FUNCNAME "(): No connection to LDAP server");
      return retval;
    }
  
  /* create data object to be sent to wait-event */
  
  w=pmalloc(p->id->p,sizeof(_waitevt));
  w->ld=ld;

  /* construct LDAP binding argument and define what attributes are to 
     be sent to the modify operation */
  
  q=cfg->vcard_attrib;
  
  for (i=0;i<size-1;i++)
    {
      if (q->name!=NULL)
	{
	  int tel=NONE,email=NONE;
	  
	  /* check for children of other named nodes */
	  
	  /* look for N node */
	  if (cw_check_n(q->name))
	    temp=xmlnode_get_tag(xmlnode_get_tag(vcard,"N"),q->name);
	  
	  /* look for ADR node */
	  else if ((adr=cw_check_adr(q->name)))
	    temp=xmlnode_get_tag(cw_get_adr_node(vcard,adr),
				 cw_tokenise_suffix(q->name));
	  
	  /* look for ORG node */
	  else if (cw_check_org(q->name))
	    temp=xmlnode_get_tag(xmlnode_get_tag(vcard,"ORG"),q->name);
	  
	  /* look for TEL node */
	  else if ((tel=cw_check_tel(q->name)))
	    temp=cw_get_tel_node(vcard,tel,
				 cw_tokenise_suffix(q->name));
	  
	  /* look for EMAIL node */
	  else if ((email=cw_check_email(q->name)))
	    temp=cw_get_email_node(vcard,email);

	  else
	    temp=xmlnode_get_tag(vcard,q->name);
	  
	  /* set up the data for the modify operation */
	  if (tel || email)
	    {
	      /* grab the telephone and email data */
	      attribs[i][0]=xmlnode_get_data(temp);
	    }

	  else
	    {
	      /* normal data get */
	      attribs[i][0]=xmlnode_get_data(xmlnode_get_firstchild(temp));
	    }
	  
	  attribs[i][1]=NULL;
      
	  /* build the modify structure */
	  modified_data[i].mod_op=LDAP_MOD_REPLACE;
	  modified_data[i].mod_type=q->mapping;
	  modified_data[i].mod_values=attribs[i];

	  modified[i]=&modified_data[i];
	}
      q=q->next;
    }
  
  /* set the last part of the LDAPMod struct to NULL */
  modified[size-1]=NULL;

  /* set up spool */
  sp=spool_new(p->id->p);
  spooler(sp,cfg->user_attrib,"=",p->id->user,",",cfg->base,sp);
  filter=spool_print(sp);
  
  log_debug(ZONE,": " FUNCNAME "(): filter: %s",filter);
  
  /* address query to the LDAP directory */
  
  if ((w->msgid=ldap_modify(w->ld,filter,modified))
      ==-1)
    {
      /* error in the modify call */
      log_error(ZONE,": " FUNCNAME "(): error in LDAP modify call");
      return retval;
    }
  
  /* wait for response to come back from LDAP directory */
  evt = pth_event(PTH_EVENT_FUNC,&js_cw_event_mon,(void *)w,
		  pth_time(0,500000));
  pth_wait(evt); /* block until LDAP responds */
  have_evt=TRUE;
  
  if (w->result==NULL)
    {
      /* an error occured during wait */
      log_error(ZONE,": " FUNCNAME "(): Error occured while waiting for result");
      pth_event_free(evt,PTH_FREE_ALL);
      return retval;
    }
  
  log_debug(ZONE,": " FUNCNAME "(): result: 0x%08x", w->result);
  
  /* check the result of the modify */
  if (w->rc!= LDAP_RES_SEARCH_ENTRY)
    {
      /* add extra checking here to distinguish the LDAP error */
      error=ldap_err2string(ldap_result2error(ld,w->result,1));
      
      if (!j_strcmp(error,"Success"))
	{
	  /* success!!! */
	  retval=r_DONE;
	}
      else
	{
	  /* failure */
	  log_error(ZONE,": " FUNCNAME "(): LDAP result \"%s\" occured during vcard modification operation",error);
	}
    }
  
  if (have_evt)
    pth_event_free(evt,PTH_FREE_ALL);
  
  log_debug(ZONE,": " FUNCNAME "():LDAP roster access completed");
  
  return retval;
}
#undef FUNCNAME




/* this function is used to build vcard data from diretory */

#define FUNCNAME "cw_build_vcard"

xmlnode cw_build_vcard_data (LDAP* ld, LDAPMessage* e, mod_ldap_conf cfg, xmlnode data)
{
  char *attribute;
  vcard q;
  BerElement *ber;

  /* build the vcard from the mappings and ldap attributes */
  for (attribute=ldap_first_attribute(ld,e,&ber);attribute!=NULL;
       attribute=ldap_next_attribute(ld,e,ber))
    {
      /* set the pointer to the vcard structure */
      q=cfg->vcard_attrib;
      
      /* check for what fields are in use */
      while (q!=NULL)
	{
	  /* check for special fields, such as N tag, TEL HOME tag etc... */
	  
	  if (j_strcmp(q->mapping,attribute)==0)
	    {
	      int tel=NONE,email=NONE,adr=NONE;
	      
	      /* add data to xmlnode */
	      char **vals=ldap_get_values(ld,e,attribute);
	      xmlnode temp=NULL;
	      
	      /* check for N node */
	      if (cw_check_n(q->name))
		{
		  /* build N node for vcard */
		  data=cw_insert_n_node(data,vals,q->name);
		}
	      
	      /* check for ORG node */
	      else if (cw_check_org(q->name))
		{
		  /* build ORG node for vcard */
		  data=cw_insert_org_node(data,vals,q->name);
		}
	      
	      /* check for ADR node */
	      else if ((adr=cw_check_adr(q->name)))
		{
		  /* build ADR node for vcard */
		  data=cw_insert_adr_node(data,vals,q->name,adr);
		}
	      
	      /* check for TEL node */
	      else if ((tel=cw_check_tel(q->name)))
		{
		  /* build TEL node for vcard */
		  data=cw_insert_tel_node(data,vals,q->name,tel);
		} 
	      
	      /* check for EMAIL node */
	      else if ((email=cw_check_email(q->name)))
		{
		  /* build EMAIL node for vcard */
		  data=cw_insert_email_node(data,vals,q->name,email);
		}
	      
	      /* node with no children */
	      else
		{
		  temp=xmlnode_new_tag(q->name);
		  xmlnode_insert_cdata(temp,vals[0],strlen(vals[0]));
		  xmlnode_insert_tag_node(data,temp);
		}
	      
	      /* free vals */
	      if (vals!=NULL)
		ldap_value_free(vals);
	    }
	  q=q->next;
	}
    }
  
  if (ber!=NULL)
    ber_free(ber,0);
  
  /* completed vcard xml node */ 
  return data;
}
#undef FUNCNAME





/* this function is used to strip down incoming xml containing vcard data
 * and build the associated LDAP modify structure */

LDAPMod** cw_strip_vcard_data (mod_ldap_conf cfg, xmlnode data, int size)
{
  char *attribs[size][2];
  LDAPMod modified_data[size],*modified[size];
  vcard q;
  int i=0;
  xmlnode vcard=xmlnode_get_firstchild(data);

  q=cfg->vcard_attrib;
  
  for (i=0;i<size-1;i++)
    {
      if (q->name!=NULL)
	{
	  /*	  char *attribs[2];*/
	  
	  xmlnode temp=xmlnode_get_tag(vcard,q->name);
	  
	  /* set up the data for the modify operation */
	  attribs[i][0]=xmlnode_get_data(xmlnode_get_firstchild(temp));
	  attribs[i][1]=NULL;
	  
	  /* build the modify structure */
	  modified_data[i].mod_op=LDAP_MOD_REPLACE;
	  modified_data[i].mod_type=q->mapping;
	  modified_data[i].mod_values=attribs[i];
	  /*  
	      if (i==0)
	      test1.mod_values=attribs[i];
	      if (i==1)
	      test2.mod_values=attribs[i];
	      if (i==2)
	      test3.mod_values=attribs[i];
	      if (i==3)
	      test4.mod_values=attribs[i];
	  */
	  modified[i]=&modified_data[i];
	}
      q=q->next;
    }
  
  /* set the last part of the LDAPMod struct to NULL */
  modified[size-1]=NULL;
  
  /* return modified data */
  /*  return modified;*/
  return NULL;
}




/* this function returns the size of the vcard linked list */

int cw_find_vcard_size(mod_ldap_conf cfg)
{
  int size=0;
  vcard q;

  q=cfg->vcard_attrib;

  while (q)
    {
      size++;
      q=q->next;
    }
  return size;
}





/* this function is used to determine if a N (name) node is present */

int cw_check_n (const char* name)
{
  /* check for N node as the following values are children
     of the N node */
  if (j_strcmp(name,"GIVEN")==0 
      || j_strcmp(name,"MIDDLE")==0
      || j_strcmp(name,"FAMILY")==0)
    return 1;
  else
    return 0;
}




/* this function is used to determine if name ORG node is present */

int cw_check_org (const char* name)
{
  /* check for ORG node as the following values are children
     of the ORG node */
  if (j_strcmp(name,"ORGNAME")==0 
      || j_strcmp(name,"ORGUNIT")==0)
    return 1;
  else
    return 0;
}
 



/* this function is used to determine if name ADR is work or home node */

int cw_check_adr (const char* name)
{
  /* check for ADR node as the following values are children
     of the ADR node */

  /* cw_build vcard_data writes vcard to client as 
     HOME->STREET->EXTADD etc... */

  if (j_strcmp(name,"STREET")==0 
      || j_strcmp(name,"EXTADD")==0
      || j_strcmp(name,"LOCALITY")==0
      || j_strcmp(name,"REGION")==0
      || j_strcmp(name,"PCODE")==0
      || j_strcmp(name,"COUNTRY")==0)
    return HOME;

  /* using suffix _WORK to differentiate between HOME and WORK
   * cw_build_vcard_data writes vcard to client as 
   * WORK->STREET->EXTADD etc...
   * and drops the suffix */
  
  else if (j_strcmp(name,"STREET_WORK")==0 
	   || j_strcmp(name,"EXTADD_WORK")==0
	   || j_strcmp(name,"LOCALITY_WORK")==0
	   || j_strcmp(name,"REGION_WORK")==0
	   || j_strcmp(name,"PCODE_WORK")==0
	   || j_strcmp(name,"COUNTRY_WORK")==0)
    return WORK;
  
  else
    return NONE;
}   




/* this function returns the correct ADR node */

xmlnode cw_get_adr_node(xmlnode vcard,int type)
{
  xmlnode temp=xmlnode_get_firstchild(vcard);
  
  /* we are looking for the node with name of ADR */

  while (xmlnode_get_name(temp)!=NULL)
    {
      if (j_strcmp(xmlnode_get_name(temp),"ADR")==0)
	{
	  /* now check children names */
	  if ((j_strcmp(xmlnode_get_name(xmlnode_get_firstchild(temp))
			,"HOME")==0 && type==HOME) ||
	      (j_strcmp(xmlnode_get_name(xmlnode_get_firstchild(temp))
			,"WORK")==0 && type==WORK))
	    {      
	      return(temp);  
	    }
	}
      temp=xmlnode_get_nextsibling(temp);
    }
  
  /* data not found, no ADR node */
  return NULL;
}




/* this function is used to check for the correct TEL node */

int cw_check_tel (const char* name)
{
  /* check for TEL node as the following values are children
     of the TEL node */

  if (j_strcmp(name,"VOICE")==0 
      || j_strcmp(name,"FAX")==0
      || j_strcmp(name,"MSG")==0)
    return HOME;

  /* using suffix _WORK to differentiate between HOME and WORK */
  
  else if (j_strcmp(name,"VOICE_WORK")==0 
	   || j_strcmp(name,"FAX_WORK")==0
	   || j_strcmp(name,"MSG_WORK")==0)
    return WORK;
  
  else
    return NONE;
}   




/* this function is used to check for the correct EMAIL node */

int cw_check_email (const char* name)
{
  /* check for EMAIL node as the following values are children
     of the EMAIL node */
  
  if (j_strcmp(name,"EMAIL")==0) 
    return PREF;
  
  else if (j_strcmp(name,"EMAIL_OTHER")==0) 
    return OTHER;
  
  else
    return NONE;
}   




/* this function returns the correct TEL node */

xmlnode cw_get_tel_node(xmlnode vcard,int type,const char* name)
{
  xmlnode temp=xmlnode_get_firstchild(vcard);
  xmlnode child=NULL;
  char* home;
  

  if (type==HOME)
    home="HOME";
  else if (type==WORK)
    home="WORK";
  else
    home="NULL";
  
  /* we are looking for the node with name of TEL */

  while (xmlnode_get_name(temp)!=NULL)
    {
      if (j_strcmp(xmlnode_get_name(temp),"TEL")==0)
	{
	  /* now check children names */
	  if (j_strcmp(xmlnode_get_name(child=xmlnode_get_firstchild(temp)),
		       name)==0)
	    {
	      /* found the right name, check for WORK or HOME */
	      
	      if (j_strcmp(xmlnode_get_name
			   (child=xmlnode_get_nextsibling(child)),
			   home)==0)
		{
		  /* found the next node. data is contained in the next
		     sibling */
		  return(xmlnode_get_nextsibling(child));
		}
	    }
	}
      
      /* go to next node */
      temp=xmlnode_get_nextsibling(temp);
    }
  
  /* data not found, no ADR node */
  return NULL;
}




/* this function returns the correct EMAIL node */

xmlnode cw_get_email_node(xmlnode vcard,int type)
{
  xmlnode temp=xmlnode_get_firstchild(vcard);
  xmlnode child=NULL;
  
  /* we are looking for the node with name of ADR */
  
  while (xmlnode_get_name(temp)!=NULL)
    {
      if (j_strcmp(xmlnode_get_name(temp),"EMAIL")==0)
	{
	  /* now check children names */
	  if (j_strcmp(xmlnode_get_name(child=xmlnode_get_firstchild(temp))
		       ,"INTERNET")==0 && type==PREF)
	    {
	      /* format is EMAIL->INTERNET->PREF->data */
	      return(xmlnode_get_nextsibling(xmlnode_get_nextsibling(child)));
	    }	  

	  else if (j_strcmp(xmlnode_get_name
			    (child=xmlnode_get_firstchild(temp))
			    ,"INTERNET")==0 && type==OTHER)
	    return(xmlnode_get_nextsibling(child));
	}
      /* get next sibling */    
      temp=xmlnode_get_nextsibling(temp);
    }
  
  /* data not found, no EMAIL node */
  return NULL;
}




/* this will insert a new N node value into the vcard from the directory */

xmlnode cw_insert_n_node (xmlnode data, char** vals, char* name)
{
  xmlnode n=NULL,temp=NULL;
  
  /* build N node for vcard */
  if ((temp=xmlnode_get_tag(data,"N"))==NULL)
    {
      /* new N node */
      temp=xmlnode_new_tag("N");
    }
  
  /* build child data for N node */
  n=xmlnode_new_tag(name);
  xmlnode_insert_cdata(n,vals[0],strlen(vals[0]));
  xmlnode_insert_tag_node(temp,n);
  
  /* insert tag into vcard */
  xmlnode_insert_tag_node(data,temp);

  /* return the new vcard data node */
  return data;
}




/* this will insert a new ORG node value into the vcard from the directory */

xmlnode cw_insert_org_node (xmlnode data, char** vals, char* name)
{
  xmlnode org=NULL,temp=NULL;
  
  /* build ORG node for vcard */
  if ((temp=xmlnode_get_tag(data,"ORG"))==NULL)
    {
      /* new ORG node */
      temp=xmlnode_new_tag("ORG");
    }
  
  /* build child data for ORG node */
  org=xmlnode_new_tag(name);
  xmlnode_insert_cdata(org,vals[0],strlen(vals[0]));
  xmlnode_insert_tag_node(temp,org);

  /* insert tag into vcard */
  xmlnode_insert_tag_node(data,temp);
  
  /* return the new vcard data node */
  return data;
}




/* this function will insert a new ADR node value into the vcard xml */

xmlnode cw_insert_adr_node(xmlnode data, char** vals, char* name, int type)
{
  xmlnode adr=NULL,temp=NULL;

  /* build ADR node for vcard */
  if ((temp=cw_get_adr_node(data,type))==NULL)
    {
      /* new ADR node */
      temp=xmlnode_new_tag("ADR");
      
      /* check type */
      if (type==WORK)
	{
	  /* WORK */
	  xmlnode_insert_tag(temp,"WORK");
	}
      else if (type==HOME)
	{
	  /* HOME */
	  xmlnode_insert_tag(temp,"HOME");
	}
    }
  
  /* build child data for ADR node */
  adr=xmlnode_new_tag(cw_tokenise_suffix(name));
  xmlnode_insert_cdata(adr,vals[0],strlen(vals[0]));
  xmlnode_insert_tag_node(temp,adr);
  
  /* insert tag into vcard */
  xmlnode_insert_tag_node(data,temp);

  /* return the new vcard data node */
  return data;  
}




/* function to insert a TEL node into vcard */

xmlnode cw_insert_tel_node (xmlnode data, char** vals, char* name, int type)
{
  xmlnode temp=NULL;
  
  /* build TEL node for vcard */
  
  temp=xmlnode_new_tag("TEL");
  
  /* insert the type of TEL contact - voice, fax or msg */
  xmlnode_insert_tag(temp,cw_tokenise_suffix(name));
  
  /* check to see if HOME or WORK */
  if (type==HOME)
    xmlnode_insert_tag(temp,"HOME");
  else if (type==WORK)
    xmlnode_insert_tag(temp,"WORK");

  /* finally, add data */
  xmlnode_insert_tag(temp,NULL);
  xmlnode_insert_cdata(temp,vals[0],strlen(vals[0]));  
  
  /* insert tag into vcard */
  xmlnode_insert_tag_node(data,temp);

  /* return the new vcard data node */
  return data;  
}




/* function to insert an EMAIL node into vcard */

xmlnode cw_insert_email_node (xmlnode data, char** vals, char* name, int type)
{
  xmlnode temp=NULL;
  
  /* build EMAIL node for vcard */
  
  temp=xmlnode_new_tag("EMAIL");
  xmlnode_insert_tag(temp,"INTERNET");

  /* check for PREF or OTHER */
  if (type==PREF)
    xmlnode_insert_tag(temp,"PREF");
  
  /* finally, add data */
  xmlnode_insert_tag(temp,NULL);
  xmlnode_insert_cdata(temp,vals[0],strlen(vals[0]));  
  
  /* insert tag into vcard */
  xmlnode_insert_tag_node(data,temp);

  /* return the new vcard data node */
  return data;  
}




/* removes the _WORK and _OTHER suffix from an entry in the vcard table */

char* cw_tokenise_suffix (char* name)
{
  char* modified_name;

  modified_name=strtok(name,"_");
  
  return (modified_name);
}










